import torch
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist 
import logging
import pandas as pd 
import traceback

from ...core import models
from ..misc import utils

class BaseManager:
    """
    Manager all modules and computation devices. Support three kinds of computation:

    1. DataParallel (single machine)
    2. DistributedDataParallel (single machine)
    3. DistributedDataParallel (multi machines)
    """
    def __init__(
        self,
        trainer,
        tester,
        recorder,
        objects_dict,
        device=None,
        schedulers=None,
        gradclipper=None,
        samplers=None,
        collatefns=None,
        is_resume=False,
        is_distributed=False,
        device_wrapper_type="DP",
        dist_port=23456,
        world_size=None,
        phase="train",
        primary_metric=["test", "recall_at_1"],
        to_device_list=["models", "collectors"],
        to_wrap_list=["models"],
        patience=10,
    ):
        self.trainer = trainer
        self.tester = tester
        self.recorder = recorder
        self.objects_dict = objects_dict
        self.device = device
        self.schedulers = schedulers
        self.gradclipper = gradclipper
        self.samplers = samplers
        self.collatefns = collatefns
        self.epochs = -1
        self.is_resume = is_resume
        self.is_distributed = is_distributed
        self.device_wrapper_type = device_wrapper_type
        self.dist_port = dist_port
        self.world_size = world_size
        self.phase = phase
        self.primary_metric = primary_metric
        self.to_device_list = to_device_list
        self.to_wrap_list = to_wrap_list
        self.patience = patience

        self.best_metric = -1
        self.patience_counts = 0
        self.is_best = False

        self.assert_phase()
        self.assert_device()
        self.assert_required_member()
        self.assert_resume_folder_exist()

        self.initiate_objects_dict()
        self.initiate_members()
    
    @property
    def _required_member(self):
        return [
            "metrics",
            "collectors",
            "selectors",
            "models",
            "losses",
            "evaluators",
            "optimizers",
            "transforms",
            "datasets",
        ]
    
    def assert_phase(self):
        assert self.phase in ["train", "evaluate"]

    def assert_device(self):
        assert self.device_wrapper_type in ["DP", "DDP"]
        if self.is_distributed:
            assert self.device_wrapper_type == "DDP"

    def assert_required_member(self):
        object_dict_keys = list(self.objects_dict.keys())
        assert all(
            [item in object_dict_keys
            for item in self._required_member]
        )
    
    def assert_resume_folder_exist(self):
        if self.is_resume:
            assert not self.recorder.delete_old_folder

    def initiate_objects_dict(self):
        for k, v in self.objects_dict.items():
            setattr(self, k, v)
        del self.objects_dict
    
    def initiate_members(self):
        self.initiate_device()
        self.initiate_models()
        self.initiate_collectors()
        self.initiate_selectors()
        self.initiate_losses()
        self.initiate_schedulers()
        # for distributed training
        if self.is_distributed:
            self.initiate_distributed_trainers()
            self.initiate_distributed_testers()
        self.initiate_addition_items()
    
    def initiate_addition_items(self):
        pass

    def initiate_device(self):
        if self.device_wrapper_type == "DDP" and not self.is_distributed:
            torch.distributed.init_process_group(
                backend='nccl',
                init_method='tcp://localhost:{}'.format(self.dist_port),
                rank=0,
                world_size=1
            )
        
        if self.is_distributed:
            self.world_size = (
                dist.get_world_size()
                if self.world_size is None
                else self.world_size
            )

        self.main_device_id, self.device_ids = None, None
        self.multi_gpu = False
        if self.device is None:
            self.main_device_id = 0
            self.device_ids = [0]
        elif isinstance(self.device, int):
            self.main_device_id = self.device
            self.device_ids = [self.device]
        elif isinstance(self.device, list):
            self.main_device_id = self.device[0]
            self.device_ids = self.device
            self.multi_gpu = (
                True if len(self.device_ids) > 1
                else False
            )
        else:
            raise TypeError(
                "Device type error!"
            )
        # initiate self.device
        self.device = torch.device(
            "cuda:{}".format(self.main_device_id)
            if torch.cuda.is_available()
            else "cpu"
        )
    
    def initiate_models(self):
        # to device
        is_to_device = "models" in self.to_device_list
        is_to_wrap = "models" in self.to_wrap_list
        if is_to_device:
            self._members_to_device("models", to_warp=is_to_wrap)
    
    def initiate_collectors(self):
        # to device
        is_to_device = "collectors" in self.to_device_list
        is_to_wrap = "collectors" in self.to_wrap_list
        if is_to_device:
            self._members_to_device("collectors", to_warp=is_to_wrap)
    
    def initiate_selectors(self):
        # to device
        is_to_device = "selectors" in self.to_device_list
        is_to_wrap = "selectors" in self.to_wrap_list
        if is_to_device:
            self._members_to_device("selectors", to_warp=is_to_wrap)

    def initiate_losses(self):
        # to device
        is_to_device = "losses" in self.to_device_list
        is_to_wrap = "losses" in self.to_wrap_list
        if is_to_device:
            self._members_to_device("losses", to_warp=is_to_wrap)
    
    def initiate_distributed_trainers(self):
        total_batch_size = self.trainer.batch_size
        assert (total_batch_size % self.world_size) == 0
        sub_batch_size = int(total_batch_size // self.world_size)
        self.trainer.set_distributed(True)
        self.trainer.set_batch_size(sub_batch_size)
    
    def initiate_distributed_testers(self):
        self.tester.set_distributed(True)
    
    def initiate_schedulers(self):
        if self.schedulers is None:
            self.schedulers = {}
    
    def _members_to_device(self, module_name: str, to_warp=True):
        members = getattr(self, module_name)
        # to device
        if not self.is_distributed:
            # single-device
            if self.multi_gpu:
                for k, v in members.items():
                    members[k] = members[k].to(self.device)
                    if to_warp:
                        if self.device_wrapper_type == "DP":
                            members[k] = torch.nn.DataParallel(
                                v,
                                device_ids=self.device_ids
                            )
                        else:
                            try:
                                members[k] = DDP(
                                    v,
                                    device_ids=self.device_ids,
                                    find_unused_parameters=True
                                )
                            except:
                                trace = traceback.format_exc()
                                logging.warning("{}".format(trace))
            else:
                for k, v in members.items():
                    members[k] = v.to(self.device)
        else:
            # multi-device
            for k, v in members.items():
                members[k] = members[k].to(self.device)
                try:
                    members[k] = DDP(
                        members[k], 
                        device_ids=self.device_ids,
                        find_unused_parameters=True
                    )
                except:
                    trace = traceback.format_exc()
                    logging.warning("{}".format(trace))
    
    """
    Run
    """
    
    def run(self, phase="train", start_epoch=0, total_epochs=61, is_test=True, is_save=True, interval=1, warm_up=2, warm_up_list=None):
        self.phase = phase
        self.assert_phase()
        self.prepare()
        self.maybe_resume(is_save=is_save)
        if self.phase == "train":
            for i in range(start_epoch, total_epochs):
                self.epochs += 1
                if i < warm_up:
                    logging.info("Warm up with {}".format(warm_up_list))
                    self.trainer.set_activated_optims(warm_up_list)
                else:
                    self.trainer.set_activated_optims()
                self.train(epochs=self.epochs)
                self.release_memory()
                if is_test:
                    if (i % interval) == 0:
                        self.test()
                        self.display_metrics()
                        self.save_metrics()
                        self.release_memory()
                if is_save:
                    self.save_models()
                # early stop
                if self.patience_counts >= self.patience:
                    logging.info("Training terminated!")
                    break
        elif self.phase == "evaluate":
            self.test()
            self.display_metrics()
    
    def prepare(self):
        # prepare trainer
        utils.func_params_mediator(
            [self],
            self.trainer.prepare
        )
        # prepare tester
        utils.func_params_mediator(
            [
                {"recorders": self.recorder},
                self,
            ],
            self.tester.prepare
        )
        
    def maybe_resume(self, is_save=True):
        if self.is_resume:
            logging.info("Resume objects...")
            self.epochs = self.recorder.load_models(
                obj=self.trainer,
                device=self.device
            )
        else:
            if is_save:
                self.save_models(is_best=True)
    
    def meta_test(self):
        self.epochs = -1
        self.test()
        self.save_metrics()
        self.display_metrics()
    
    def save_metrics(self):
        for k, v in self.metrics.items():
            data, _ = self.recorder.get_data({k:v})
            self.recorder.update(data, self.epochs)
    
    def display_metrics(self):
        # best metric check
        cur_metric = self.metrics[self.primary_metric[0]][self.primary_metric[1]]
        if cur_metric > self.best_metric:
            self.best_metric = cur_metric
            self.is_best = True
            logging.info("NEW BEST METRIC!!!")
            self.patience_counts = 0
        else:
            self.is_best = False
            self.patience_counts += 1
        self.metrics[self.primary_metric[0]]["BEST_" + self.primary_metric[1]] = self.best_metric

        # display
        for k, v in self.metrics.items():
            logging.info("{} Metrics ---".format(k.upper()))
            print(pd.DataFrame([v]))
        
    def save_models(self, is_best=None):
        if is_best is None:
            is_best = self.is_best
        self.recorder.save_models(self.trainer, step=self.epochs, best=is_best)
    
    def train(self, epochs=None):
        self.trainer.train(epochs=epochs)
    
    def test(self):
        self.metrics = self.tester.test()
    
    def release_memory(self):
        torch.cuda.empty_cache()



import torch
import numpy as np 
from torchvision import datasets
import os 

from .base_dataset import BaseDataset

class MiniImageNet(BaseDataset):
    def init_dataset(self):
        self.root = os.path.join(self.root, "mini_imagenet")
        img_folder = os.path.join(self.root, self.phase)
        self.dataset = datasets.ImageFolder(img_folder)
        self.labels = np.array([b for (a,b) in self.dataset.imgs])
        assert len(np.unique(self.labels)) == self.assert_num_classes
        assert self.__len__() == self.assert_num_samples
    
    def get_labels(self):
        return self.labels
    
    def __len__(self):
        return len(self.dataset)
    
    def get_image_label(self, idx):
        return self.dataset[idx]



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class TripletLoss(BaseLoss):
    """
    paper: `Learning local feature descriptors with triplets and shallow convolutional neural networks <https://www.researchgate.net/profile/Krystian_Mikolajczyk/publication/317192886_Learning_local_feature_descriptors_with_triplets_and_shallow_convolutional_neural_networks/links/5a038dad0f7e9beb1770c3c2/Learning-local-feature-descriptors-with-triplets-and-shallow-convolutional-neural-networks.pdf>`_
    """
    def __init__(
        self,
        margin=1,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.margin = margin
    
    def required_metric(self):
        return ["euclid"]

    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        is_same_source=False,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        pos_pair, neg_pair = l_f.indices_to_pairs(metric_mat, indices_tuple, assert_shape=[3])

        triplet_loss = torch.nn.functional.relu(pos_pair - neg_pair + self.margin)

        # mean_triplet_loss = torch.mean(triplet_loss)

        nonzero_triplet_loss = triplet_loss[torch.where(triplet_loss)[0]]
        loss = torch.mean(nonzero_triplet_loss)
        return loss

    



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class CircleLoss(BaseLoss):
    """
    modified from: https://github.com/KevinMusgrave/pytorch-metric-learning

    paper: `Circle Loss: A Unified Perspective of Pair Similarity Optimization <http://openaccess.thecvf.com/content_CVPR_2020/html/Sun_Circle_Loss_A_Unified_Perspective_of_Pair_Similarity_Optimization_CVPR_2020_paper.html>`_
    """
    def __init__(
        self,
        m=0.4,
        gamma=80,
        **kwargs
    ):
        super(CircleLoss, self).__init__(**kwargs)
        self.m = m
        self.gamma = gamma
        self.op = 1 + m
        self.on = - m
        self.delta_p = 1 - m
        self.delta_n = m
    
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        pos_mask = (row_labels == col_labels)
        neg_mask = ~ pos_mask
        if is_same_source:
            pos_mask.fill_diagonal_(False)
        
        tmp_mat = torch.zeros_like(metric_mat)
        pos_pair = metric_mat[pos_mask]
        neg_pair = metric_mat[neg_mask]

        # construct matrix
        tmp_mat[pos_mask] = (
            - self.gamma
            * torch.relu(self.op - pos_pair.detach())
            * (pos_pair - self.delta_p)
        )
        tmp_mat[neg_mask] = (
            self.gamma
            * torch.relu(neg_pair.detach() - self.on)
            * (neg_pair - self.delta_n)
        )

        # compute logsumexp
        se_pos = l_f.sumexp(
            tmp_mat, keep_mask=pos_mask, dim=1
        )
        se_neg = l_f.sumexp(
            tmp_mat, keep_mask=neg_mask, dim=1
        )

        loss = torch.log(1 + se_pos * se_neg)
        zero_rows = torch.where(
            (torch.sum(pos_mask, dim=1) != 0) & 
            (torch.sum(neg_mask, dim=1) != 0)
        )[0]
        loss = loss[zero_rows]
        if len(loss) == 0:
            loss = torch.sum(metric_mat * 0)
        else:
            loss = torch.mean(loss)
        return loss




import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class PosPairLoss(BaseLoss):
    """
    Designed for SimSiam.

    paper: `Exploring Simple Siamese Representation Learning <https://arxiv.org/abs/2011.10566>`_
    """
    def __init__(
        self,
        **kwargs
    ):
        super(PosPairLoss, self).__init__(**kwargs)
    
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        pos_mask = (row_labels == col_labels)
        pos_pair = metric_mat[pos_mask]
        loss = torch.mean( - pos_pair)
        return loss


import pretrainedmodels as ptm
import torch
import torch.nn as nn 
import logging

# Code from: Pytorch Metric BaseLine
class resnet50(nn.Module):
    """
    Container for ResNet50 s.t. it can be used for metric learning.
    The Network has been broken down to allow for higher modularity, if one wishes
    to target specific layers/blocks directly.
    """
    def __init__(self, pretrained=True, list_style=False, no_norm=False):
        super(resnet50, self).__init__()

        if pretrained:
            logging.info('Getting pretrained weights...')
            self.model = ptm.__dict__['resnet50'](num_classes=1000, pretrained='imagenet')
        else:
            logging.info('Not utilizing pretrained weights!')
            self.model = ptm.__dict__['resnet50'](num_classes=1000, pretrained=None)

        for module in filter(lambda m: type(m) == nn.BatchNorm2d, self.model.modules()):
            module.eval()
            module.train = lambda _: None
            
        # self.last_linear = self.model.last_linear
        self.last_linear = self.model.last_linear
        self.model.last_linear = torch.nn.Linear(self.model.last_linear.in_features, 128)

        self.layer_blocks = nn.ModuleList([self.model.layer1, self.model.layer2, self.model.layer3, self.model.layer4])

    def forward(self, x, is_init_cluster_generation=False):
        x = self.model.maxpool(self.model.relu(self.model.bn1(self.model.conv1(x))))

        for layerblock in self.layer_blocks:
            x = layerblock(x)

        x = self.model.avgpool(x)
        x = x.view(x.size(0),-1)
        return x



import numpy as np 
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

class HardPairSelector(BaseSelector):
    """
    A self-defined selector according to distance weighted sampling method
    """
    def __init__(self, hardneg_cutoff=0.5, is_similarity=False, **kwargs):
        super(HardPairSelector, self).__init__(**kwargs)
        self.hardneg_cutoff = hardneg_cutoff
        self.is_similarity = is_similarity
    
    def forward(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False
    ) -> tuple:
        bs = metric_mat.size(0)
        device = metric_mat.device
        dtype = metric_mat.dtype
        # pos and neg mask
        matches = (row_labels == col_labels).byte()
        diffs = matches ^ 1

        if not self.is_similarity:
            hardneg_cutoff = min(
                torch.max(metric_mat[matches.bool()]).item(), self.hardneg_cutoff
            )
            metric_mat_to_weights = metric_mat.clamp(min=hardneg_cutoff).detach()
        else:
            hardneg_cutoff = max(
                torch.min(metric_mat[matches.bool()]).item(), self.hardneg_cutoff
            )
            metric_mat_to_weights = metric_mat.clamp(max=hardneg_cutoff).detach()
        

        # construct the tuple
        if is_same_source:
            matches.fill_diagonal_(0)
        
        has_pos_mask = torch.where(
            torch.sum(matches, dim=-1) > 0
        )[0]

        ## anchor indices
        a_ids = torch.arange(bs).to(device)
        ## positive indices
        p_ids = torch.multinomial(
            input=matches[has_pos_mask].float(),
            num_samples=1,
            replacement=True
        ).flatten()
        pos_pairs = torch.stack([a_ids[has_pos_mask], p_ids], dim=1)
        pos_flags = torch.ones(pos_pairs.shape[0], 1)

        ## negative indices
        if not self.is_similarity:
            metric_mat_to_weights[~diffs.bool()] = torch.finfo(dtype).max
            n_ids = torch.argmin(
                metric_mat_to_weights, dim=-1
            )
        else:
            metric_mat_to_weights[~diffs.bool()] = torch.finfo(dtype).min
            n_ids = torch.argmax(
                metric_mat_to_weights, dim=-1
            )
        neg_pairs = torch.stack([a_ids, n_ids], dim=1)
        neg_flags = torch.zeros(neg_pairs.shape[0], 1)
        tuples = torch.cat([pos_pairs, neg_pairs], dim=0)
        flags = torch.cat([pos_flags, neg_flags], dim=0).byte()

        indices_tuple = {
            INDICES_TUPLE: tuples,
            INDICES_FLAG: flags
        }
        weight = None
        return (
            metric_mat,
            row_labels,
            col_labels,
            is_same_source,
            indices_tuple,
            weight,
        )



import torch

from .base_similarity_metric import BaseSimilarityMetric

class MoCoMetric(BaseSimilarityMetric):
    def __call__(self, queries, keys, queue):
        # positive matrix: Nx1
        pos_mat = torch.einsum("ij,ij->i", [queries, keys]).unsqueeze(-1)
        # negative matrix: NxK
        neg_mat = torch.einsum("ik,kj->ij", [queries, queue.t()])
        # concat
        matrix = torch.cat([pos_mat, neg_mat], dim=1)
        return matrix
         


import torch
import torch.nn.functional as F 
from ..base_loss import BaseLoss

class CrossEntropyLoss(BaseLoss):
    """
    paper: `Momentum Contrast for Unsupervised Visual Representation Learning <http://openaccess.thecvf.com/content_CVPR_2020/html/He_Momentum_Contrast_for_Unsupervised_Visual_Representation_Learning_CVPR_2020_paper.html>`_

    Cross-entropy loss designed for MoCo.
    """
    def __init__(
        self,
        **kwargs
    ):
        super().__init__(**kwargs)
    
    def required_metric(self):
        return ["moco"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False,
        *args,
        **kwargs
    ) -> torch.Tensor:
        dtype, device = metric_mat.dtype, metric_mat.device
        pos_mask = (row_labels == col_labels).type(dtype).to(device)
        pos_index = torch.where(pos_mask)[1]

        loss = F.cross_entropy(metric_mat, pos_index)
        return loss


from .base_creator import BaseCreator
import logging
from ....core import transforms
from ...misc import utils

class transformsCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [None]
    
    def create(self, module_type, module_args, *args, **kwargs):
        """
        func: create() is specially overrided for instantiating transform objects.
        """
        logging.info("+++ {}'s params: \n{}".format(
            module_type, self.print_params(module_args)
        ))
        # get model_properties
        model = module_args["model_properties"]
        params = module_args["compose_list"]
        try:
            model_properties = {
                k: getattr(model, k) 
                for k in ["mean", "std", "input_space", "input_range"]
            }
        except (KeyError, AttributeError):
            model_properties = {
                "mean": [0.485, 0.456, 0.406],
                "std": [0.229, 0.224, 0.225]
            }
        # get addition_params 
        addition_params = []
        # input_space
        input_space = model_properties.get("input_space", None)
        if input_space == "BGR":
            addition_params.append({
                "ConvertToBGR": {}
            })
        # to tensor
        addition_params.append({
            "ToTensor": {}
        })
        # input_range
        input_range = model_properties.get("input_range", None)
        if isinstance(input_range, list)  and input_range[1] != 1:
            addition_params.append({
                "Multiplier": {
                    "multiple": input_range[1]
                }
            })
        # mean, std
        mean = model_properties.get("mean", None)
        std = model_properties.get("std", None)
        if None not in [mean, std]:
            addition_params.append({
                "Normalize": {
                    "mean": mean,
                    "std": std
                }
            })
        
        # create transform.Compose
        params += addition_params
        compose_list = [
            getattr(
                transforms,
                utils.get_first_dict_key(item),
                None
            )(**utils.get_first_dict_value(item))
            for item in params
        ]
        module_object = transforms.Compose(compose_list)

        # whether to wrap
        wrapper_name = module_args.get("wrapper", False)
        if wrapper_name:
            wrapper = getattr(
                transforms,
                wrapper_name
            )
            module_object = wrapper(module_object)
        return module_object



import numpy as np
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

def get_all_triplets_indices(row_labels, col_labels=None, is_same_source=False):
    if col_labels is None:
        col_labels = row_labels.t()
    matches = (row_labels == col_labels).byte()
    diffs = matches ^ 1
    if is_same_source:
        matches.fill_diagonal_(0)
    triplets = matches.unsqueeze(2) * diffs.unsqueeze(1)
    return torch.stack(torch.where(triplets), dim=1)


class DenseTripletSelector(BaseSelector):   
    """
    Select all triplets.
    """ 
    def forward(
        self, 
        metric_mat, 
        row_labels, 
        col_labels, 
        is_same_source=False
    ) -> tuple:
        """
        Select all triplets.
        """
        weight = None
        tuples = get_all_triplets_indices(row_labels, col_labels, is_same_source)
        indices_tuple = {INDICES_TUPLE:tuples, INDICES_FLAG:None}
        return metric_mat, row_labels, col_labels, is_same_source, indices_tuple, weight


import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class MarginLoss(BaseLoss):
    """
    paper: `Sampling Matters in Deep Embedding Learning <http://openaccess.thecvf.com/content_iccv_2017/html/Wu_Sampling_Matters_in_ICCV_2017_paper.html>`_
    """
    def __init__(
        self,
        alpha=0.2,
        beta=1.2,
        nu=0,
        num_classes=100,
        beta_constant=False,
        is_similarity=False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.nu = nu
        self.num_classes = num_classes
        self.is_similarity = is_similarity
        self.beta_constant = beta_constant
        self.beta = (
            beta if beta_constant
            else torch.nn.Parameter(torch.ones(num_classes) * beta)
        )
    
    def required_metric(self):
        return ["euclid"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        is_same_source=False,
        *args,
        **kwargs
    ) -> torch.Tensor:
        a1, p, a2, n = l_f.split_indices(indices_tuple)
        pos_pair, neg_pair = l_f.indices_to_pairs(metric_mat, indices_tuple)

        if self.beta_constant:
            beta_pos = beta_neg = self.beta
        else:
            beta_pos = self.beta[row_labels[a1].flatten()]
            beta_neg = self.beta[row_labels[a2].flatten()]

        if not self.is_similarity:
            pos_loss = torch.nn.functional.relu(self.alpha + pos_pair - beta_pos)
            neg_loss = torch.nn.functional.relu(self.alpha - neg_pair + beta_neg)
        else:
            pos_loss = torch.nn.functional.relu(beta_pos + self.alpha - pos_pair)
            neg_loss = torch.nn.functional.relu(neg_pair - beta_neg + self.alpha)

        # mean_triplet_loss = torch.mean(triplet_loss)

        pos_loss = pos_loss[torch.where(pos_loss)[0]]
        neg_loss = neg_loss[torch.where(neg_loss)[0]]

        numerator = torch.sum(pos_loss) + torch.sum(neg_loss)
        denominator = len(pos_loss) + len(neg_loss)
        loss = numerator / denominator
        return loss


import torch
import numpy as np 

class BaseCollateFn(object):
    def __init__(self):
        pass

    def __call__(self, info_dict: list):
        assert isinstance(info_dict, list)
        elem = info_dict[0]
        assert isinstance(elem, dict)
        info_dict = {
            key: self._concatenate_tensor(
                [item[key] for item in info_dict]
            )
            for key in elem
        }
        info_dict = self.modify_info_dict(info_dict)
        return info_dict
    
    def modify_info_dict(self, info_dict: dict) -> dict:
        info_dict.pop("id")
        return info_dict
    
    def _concatenate_tensor(self, batch: torch.Tensor):
        elem = batch[0]
        out = None
        if isinstance(elem, torch.Tensor):
            if torch.utils.data.get_worker_info() is not None:
                # modified from: https://github.com/pytorch/pytorch/torch/utils/data/_utils/collate.py
                numel = sum([x.numel() for x in batch])
                storage = elem.storage()._new_shared(numel) # TODO: What's Tensor.FloatStorage? 
                out = elem.new(storage)
            return torch.stack(batch, 0, out=out)
        elif isinstance(elem, int):
            return torch.tensor(batch)
        elif isinstance(elem, np.int64):
            return torch.tensor(batch)
        else:
            raise TypeError("Wrong type {}".format(type(elem)))

# MoCo collate fn
class MoCoCollateFn(BaseCollateFn):
    def __init__(self):
        super(MoCoCollateFn, self).__init__()

    def modify_info_dict(self, info_dict):
        return info_dict # TODO:

# SimSiam collate fn
class SimSiamCollateFn(BaseCollateFn):
    def __init__(self):
        super(SimSiamCollateFn, self).__init__()

    def modify_info_dict(self, info_dict):
        info_dict.pop("id")
        # generate self-supervise labels
        data_len = len(info_dict["data"])
        info_dict["labels"] = torch.arange(data_len).repeat(1, 2).flatten()
        # concatenate two streams
        data_stream2 = info_dict.pop("addition_data")
        info_dict["data"] = torch.cat([info_dict["data"], data_stream2], dim=0)
        return info_dict



import torchvision.transforms.functional as F
from PIL import Image

class TwoCropsTransformWrapper(object):
    """
    Take two random crops of one image as the query and key.
    modified from: https://github.com/facebookresearch/moco
    """
    def __init__(self, base_transform):
        self.base_transform = base_transform
    
    def __call__(self, img):
        query = self.base_transform(img)
        key = self.base_transform(img)
        return {
            "data": query,
            "addition_data": key
        }

# class CombineCropsTransformWrapper(object):
#     def __init__(self, base_transform):
#         self.base_transform = base_transform
    
#     def __call__(self, img):
#         stream1 = self.base_transform(img)
#         stream2 = self.base_transform(img)
#         return {
#             "data": # TODO: 
#         }

class DefaultTransformWrapper(object):
    """
    Default wrapper.
    """
    def __init__(self, base_transform):
        self.base_transform = base_transform
    
    def __call__(self, img):
        img = self.base_transform(img)
        return {
            "data": img,
            "addition_data": img
        }


import os
import logging
import csv
import yaml
import wandb
from copy import deepcopy
from torch.utils.tensorboard import SummaryWriter
import torch
from . import utils
from ..config.setting.recorder_setting import CSV_FOLDER, MODEL_FOLDER, BOARD_FOLDER, WANDB_FOLDER, TO_SAVE_LIST, DEVICE, STEP_MODEL_SUFFIX, BEST_MODEL_SUFFIX, LOAD_EXCEPT_FUNC

class BaseInfoWriter:
    """
    This writer will save information into csv files and tensorboard. And manage model saving and loading.
        - root
        - csv
        - model
        - board
    """
    def __init__(
        self,
        project_name,
        root,
        exp_name,
        delete_old_folder=True,
        hint_if_exist=True,
        is_resume=False,
        use_wandb=True,
        params_to_save=None,
        link_config=None,
        pipeline_to_save=None,
    ): 
        self.project_name = project_name
        self.root = root
        self.exp_name = exp_name
        self.delete_old_folder = delete_old_folder
        self.hint_if_exist = hint_if_exist
        self.is_resume = is_resume
        self.use_wandb = use_wandb
        self.params_to_save = params_to_save
        self.link_config = link_config
        self.pipeline_to_save = pipeline_to_save

        self.csv_path, self.model_path, self.board_path, self.wandb_path = None, None, None, None

        self.folders_list = ['csv', 'model', 'board'] # can be modified
        self.update_ordered_list = ['csv', 'board'] # can be modified
        self.update_function_prefix = '_update_'
        self.init_ordered_list = ['board', 'model']
        self.init_function_prefix = '_init_'

        if self.use_wandb:
            self.folders_list += ['wandb']
            self.update_ordered_list += ['wandb']
            self.init_ordered_list += ['wandb']

        self.init_folders()
        self.init_update_handler()
    
    """
    INITIATE
    """

    def _meta_path_factory(self, item):
        path = os.path.join(self.root, globals()["{}_FOLDER".format(item.upper())])
        setattr(self, "{}_path".format(item), path)
        utils.create_folder(path, is_resume=self.is_resume, delete_old_folders=self.delete_old_folder)
        return None
    
    def init_folders(self):
        # create root folder
        self.root = utils.create_folder(self.root, is_resume=self.is_resume, hint_if_exist=self.hint_if_exist, delete_old_folders=self.delete_old_folder)
        # save config file
        self.save_params()
        # save pipeline flow chart
        self.save_pipeline()
        # create sub folders
        for item in self.folders_list:
            self._meta_path_factory(item)
    
    def save_params(self):
        if isinstance(self.params_to_save, dict):
            param_path = os.path.join(self.root, "param")
            utils.create_folder(param_path, hint_if_exist=False, delete_old_folders=True)
            for class_name, item_list in self.link_config.items():
                sub_param_path = os.path.join(param_path, class_name)
                os.makedirs(sub_param_path)
                for item_info in item_list:
                    # get yaml file name
                    item_name = utils.get_first_key_of_dict(item_info)
                    yaml_name = item_info[item_name]
                    sub_save_path = os.path.join(sub_param_path, item_name + "-" + yaml_name)
                    # get dictionary to save
                    sub_dict = deepcopy(self.params_to_save[class_name][item_name])
                    object_name = sub_dict.pop("type")
                    sub_dict = {object_name: sub_dict}
                    # save
                    with open(sub_save_path, mode="w", encoding="utf-8") as f:
                        yaml.dump(sub_dict, f, allow_unicode=True)
    
    def save_pipeline(self):
        if self.pipeline_to_save is not None:
            try:
                from graphviz import Digraph
                file_path = os.path.join(self.root, "pipeline")
                gz = Digraph(
                    name="pipeline",
                    node_attr={"style": "rounded", "shape": "box"}
                )
                for start, end, note in self.pipeline_to_save:
                    gz.edge(start, end, note)
                gz.render(file_path)
                logging.info("Pipeline flow chart is stored in {}".format(file_path))
            except:
                logging.warn("GraphViz isn't installed! Pipeline flow chart generation FAILED!")

    
    def init_update_handler(self):
        utils.meta_call_factory(
            obj=self,
            ordered_list=self.init_ordered_list,
            prefix=self.init_function_prefix,
            description_str="Initiate update handler by {}"
        )

    def _init_board(self):
        self.board_handler = SummaryWriter(log_dir=self.board_path)
    
    def _init_model(self):
        self.model_handler = BaseModelHandler(self.model_path)
        self.save_models = self.model_handler.save_models
        self.load_models = self.model_handler.load_models
    
    def _init_wandb(self):
        wandb.init(
            project=self.project_name,
            name=self.exp_name,
            dir=self.wandb_path,
            settings=wandb.Settings(start_method="fork")
        )
    
    def log_config(self, config_dict):
        if self.use_wandb:
            wandb.config.update(config_dict)
            logging.info("Wandb config updated!")
        else:
            logging.info("NOT use wandb!")
    
    """
    UPDATE 
    """

    def update(self, data: dict, step: int=None):
        """
        This is the primary API for BaseInfoWriter class.

        Args:
            data        (dict):     Dict
            step        (int):      Step (optional)
        """
        utils.meta_call_factory(
            obj=self,
            ordered_list=self.update_ordered_list,
            prefix=self.update_function_prefix,
            data=data,
            step=step
        )
    
    def _update_csv(self, data, step=None):
        table_name = utils.get_first_key_of_dict(data)
        table_name_csv = table_name + '.csv'
        table_path = os.path.join(self.csv_path, table_name_csv)
        data[table_name]['step'] = step
        if not os.path.exists(table_path):
            with open(table_path, mode='w', newline='') as f:
                f_csv = csv.DictWriter(f, fieldnames=list(data[table_name].keys()))
                f_csv.writeheader()
                f_csv.writerow(data[table_name])
        else:
            with open(table_path, mode='a', newline='') as f:
                f_csv = csv.DictWriter(f, fieldnames=list(data[table_name].keys()))
                f_csv.writerow(data[table_name])
        data[table_name].pop('step')

    def _update_board(self, data, step=None):
        primary_key = utils.get_first_key_of_dict(data)
        for k, v in data[primary_key].items():
            full_key = primary_key + '/' + k
            self.board_handler.add_scalar(full_key, utils.get_value(v), global_step=step)
    
    def _update_wandb(self, data, step=None):
        primary_key = utils.get_first_key_of_dict(data)
        wandb.log(data={primary_key+'/step': step})
        for k, v in data[primary_key].items():
            full_key = primary_key + '/' + k
            wandb.log(data={full_key: utils.get_value(v)})
    
    def __del__(self):
        wandb.finish()

class BaseModelHandler:
    def __init__(self, model_path):
        self.model_path = model_path

    def delete_models(self):
        models_list = os.listdir(self.model_path)
        for model_name in models_list:
            if 'best' not in model_name:
                model_item_path = os.path.join(self.model_path, model_name)
                os.remove(model_item_path)
    
    def save_models(self, obj, step, best=False, delete_old=True):
        """
        Args:
            best (bool)
        """
        if delete_old: self.delete_models()
        to_save_list = getattr(obj, TO_SAVE_LIST, {})
        for to_save_name in to_save_list:
            to_save_item = getattr(obj, to_save_name, {})
            assert isinstance(to_save_item, dict), "{} must a dictionary!".format(to_save_name)
            for k, v in to_save_item.items():
                curr_model_name = to_save_name + "_" + k + STEP_MODEL_SUFFIX.format(int(step))
                curr_model_path = os.path.join(self.model_path, curr_model_name)
                torch.save(v.state_dict(), curr_model_path)
                logging.info('{} is saved in {}'.format(curr_model_name, curr_model_path))
                if best:
                    best_model_name = to_save_name + "_" + k + BEST_MODEL_SUFFIX
                    best_model_path = os.path.join(self.model_path, best_model_name)
                    torch.save(v.state_dict(), best_model_path)
                    logging.info('{} is saved in {}'.format(best_model_name, best_model_path))
    
    def load_models(self, obj, step=None, device=None, model_path=None):
        """
        TODO: resume from the checkpoint with max number if "best" models doesn't exist.
        """
        model_path = (
            self.model_path
            if model_path is None
            else model_path
        )
        max_index_list = [-2]
        to_save_list = getattr(obj, TO_SAVE_LIST, {})
        device = getattr(obj, DEVICE, torch.device('cpu')) if device is None else device
        for to_save_name in to_save_list:
            to_save_item = getattr(obj, to_save_name, {})
            for k, v in to_save_item.items():
                import glob
                item_search = glob.glob(os.path.join(model_path, to_save_name + "_" + k+"*"))
                try:
                    index_list = [item.split(".")[0].split("_")[-1] for item in item_search]
                    index_list = [int(item) if item!="best" else -2 for item in index_list]
                    max_index = max(index_list)
                except:
                    logging.warn("Can't recognize the last training epoch!")
                    max_index = -2
                max_index_list.append(max_index)
                if len(item_search) < 1:
                    logging.warn("{} does not exist!".format(model_path))
                curr_model_name = (
                    to_save_name + "_" + k + STEP_MODEL_SUFFIX.format(int(step)) 
                    if step is not None else 
                    to_save_name + "_" + k + BEST_MODEL_SUFFIX
                )
                curr_model_path = os.path.join(model_path, curr_model_name)
                if curr_model_path not in item_search:
                    curr_model_path = item_search[0]
                try:
                    v.load_state_dict(torch.load(curr_model_path, map_location=device))
                except:
                    state_dict = torch.load(curr_model_path, map_location=device)
                    from collections import OrderedDict
                    new_state_dict = OrderedDict()
                    for p_k, p_v in state_dict.items():
                        name = LOAD_EXCEPT_FUNC(p_k)
                        new_state_dict[name] = p_v
                    v.load_state_dict(new_state_dict)
                    v = v.to(device)
                logging.info('{} is loaded from {}'.format(curr_model_name, curr_model_path))
        return max(max_index_list)



    


import os
import logging
from .creator_manager import CreatorManager
from ..misc import utils
from ...config.setting.launcher_setting import (
    CLASS_KEY,
    PARAMS_KEY,
    INITIATE_KEY,
    WRAPPER_KEY,
    OBJECTS_KEY,
    SEARCH_PARAMS_CONDITION,
    SEARCH_PARAMS_FLAG,
    SEARCH_TOP_CLASS,
    SEARCH_TAR_NAME,
    WILDCARD,
    SPLITER,
    ATTRSYM,
    INITATE_ORDER_LIST,
    WRAPPER_LIST,
    LINK_GENERAL_KEY,
    PIPELINE_KEY,
)

class ConfigHandler:
    """
    This class takes charge of reading yaml files, combining config parameters, modifying local parameters and initializing modules.

    ``ConfigHandler`` will call ``CreatorManager`` to use corresponding sub-creator to do the initialization respectly.

    Args:
        link_path (str):
            Path where the ``link.yaml`` is stored.
        params_path (str):
            Path where the params configs are stored.
        assert_path (str):
            Path where the ``assert.yaml`` is stored. (To be done!)
        is_confirm_first (bool):
            Whether to confirm before starting initialization. default: True.

    Example:
        >>> config_handler = ConfigHandler()
        >>> config_handler.get_params_dict()
        >>> objects_dict = config_handler.create_all()

    Todo:
        - assert parameters.
    """
    def __init__(
        self,
        convert_dict={},
        link_path=None,
        assert_path=None,
        params_path=None,
        wrapper_path=None,
        is_confirm_first=True,
    ):
        self.convert_dict = convert_dict
        self.link_path = link_path
        self.assert_path = assert_path
        self.params_path = params_path
        self.wrapper_path = wrapper_path
        self.is_confirm_first = is_confirm_first

        self.creator_manager = CreatorManager()
        self.initiate_params()
    
    @property
    def core_modules_order_to_create(self):
        return INITATE_ORDER_LIST
    
    @property
    def output_wrapper_list(self):
        return WRAPPER_LIST

    def _get_params_dict_operation(self, data_info, module_type):
        """
        Parse params_dict and wrapper_dict recursively.
        """
        assert isinstance(data_info, list)
        self.params_dict[module_type] = {}
        if module_type in self.output_wrapper_list:
            self.wrapper_dict[module_type] = {}
        for item in data_info:
            # parameters
            item_key = utils.get_first_dict_key(item)
            item_value = utils.get_first_dict_value(item)
            item_params_path = os.path.join(self.params_path, module_type, item_value)
            item_params_dict = utils.load_yaml(item_params_path)
            class_value = utils.get_first_dict_key(item_params_dict)
            params_value = item_params_dict[class_value][PARAMS_KEY]
            initiate_value = item_params_dict[class_value].get(INITIATE_KEY, None)
            self.params_dict[module_type][item_key] = {
                CLASS_KEY: class_value,
                PARAMS_KEY: params_value,
                INITIATE_KEY: initiate_value
            }
            # wrapper
            if module_type in self.output_wrapper_list:
                item_wrapper_path = os.path.join(self.wrapper_path, module_type, class_value + ".yaml")
                try:
                    item_wrapper_dict = utils.load_yaml(item_wrapper_path)
                    logging.info("Load specific wrapper config: {}".format(item_wrapper_path))
                except:
                    item_wrapper_path = os.path.join(self.wrapper_path, module_type, "_DEFAULT.yaml")
                    item_wrapper_dict = utils.load_yaml(item_wrapper_path)
                    logging.info("Load default wrapper config: {}".format(item_wrapper_path))
                # search pipeline
                new_map = {}
                for k, v in item_wrapper_dict["map"].items():
                    query_key = "/".join([module_type, item_key, k])
                    target_route = None
                    for pipe_k, pipe_v in self.pipeline_setting.items():
                        # TODO: for multi output 
                        # for single output
                        if query_key in pipe_k:
                            target_route = pipe_v
                    assert target_route is not None, "NO target-route matched with {}".format(query_key)
                    new_k = target_route
                    new_map[new_k] = v
                item_wrapper_dict["map"] = new_map
                self.wrapper_dict[module_type][item_key] = item_wrapper_dict
    
    """
    config setting
    """
    def register_packages(self, module_name, extra_package):
        """
        Register new packages into the specific module-creator.

        Args:
            module_name (str):
                The specific module-creator.
            extra_package (list or module):
                Extra packages to be added.
        """
        self.creator_manager.register_packages(module_name, extra_package)
    
    """
    About construction of objects dict
    """

    def _get_objects_dict_operation(self, params_info, module_type):
        assert isinstance(params_info, dict) 
        output = {}
        for k, v in params_info.items():
            self._maybe_search_params(
                module_params=v,
                instance_name=k
            )
            output[k] = self.creator_manager.create(
                module_type=module_type,
                module_params=v,
            )
            logging.info(
                "... {}: {} created, id={}".format(
                    k,
                    v[CLASS_KEY],
                    id(output[k])
                )
            )
        return output
    
    def _maybe_search_params(self, module_params, instance_name):
        module_args = module_params[PARAMS_KEY]
        if isinstance(module_args, dict):
            for k, v in module_args.items():
                search_func_name = SEARCH_PARAMS_CONDITION(v)
                if search_func_name:
                    top_class = SEARCH_TOP_CLASS(v)
                    target_name = SEARCH_TAR_NAME(v)
                    search_func_name = search_func_name.replace(SEARCH_PARAMS_FLAG, "").lower()
                    module_args[k] = getattr(self, search_func_name)(
                        top_class=top_class, 
                        instance_name=instance_name, 
                        target_name=target_name
                    )
        return module_args
    
    def _search_with_same_name_(self, top_class, instance_name, target_name):
        instance_dict = utils.operate_dict_recursively(
            src_dict=self.objects_dict[top_class],
            condition=lambda k, v: v==instance_name,
            operation=lambda x, params: x
        )
        return instance_dict[instance_name]
    
    def _search_with_target_name_(self, top_class, instance_name, target_name):
        instance_dict = utils.operate_dict_recursively(
            src_dict=self.objects_dict[top_class],
            condition=lambda k, v: v==target_name,
            operation=lambda x, params: x
        )
        return instance_dict[target_name]
    
    def _search_with_target_attr_(self, top_class, instance_name, target_name):
        module_name, attr_name = target_name.split("/")
        instance_dict = utils.operate_dict_recursively(
            src_dict=self.objects_dict[top_class],
            condition=lambda k, v: v==module_name,
            operation=lambda x, params: x
        )
        return getattr(instance_dict[module_name], attr_name)
    
    def _pass_with_named_member_(self, top_class, instance_name, target_name):
        return getattr(
            self, target_name
        )
    
    """
    About construction of link config and params dict
    """
    
    def get_params_dict(self, link_config=None, modify_link_dict=None):
        """
        Read and combine config parameters.

        Args:
            link_config (dict):
                Link dictionary.
            modify_link_dict (dict):
                Modify link config. (Default = None)

        Returns:
            dict: params' dictionary.
        """
        link_config = (
            self.link_config
            if link_config is None
            else link_config
        )

        # maybe modify link dict
        if modify_link_dict is not None:
            for k, v in modify_link_dict.items():
                assert isinstance(v, list)
                for item in v:
                    key = utils.get_first_dict_key(item)
                    value = utils.get_first_dict_value(item)
                    index = -1
                    target_list = link_config.get(k, [])
                    for idx, t_item in enumerate(target_list):
                        if utils.get_first_dict_key(t_item) == key:
                            index = idx
                    if index == -1:
                        link_config[k] = target_list
                        link_config[k].append(
                            {key: value}
                        )
                        logging.info("Add {} to {}".format(key, value))
                    else:
                        link_config[k][index] = {key: value}
                        logging.info("Modify {} to {}".format(key, value))
            logging.info("Link-config has been modified!")
        
        self.params_dict = {}
        self.wrapper_dict = {}
        self.pipeline_to_save = []
        for k, v in link_config.items():
            self._get_params_dict_operation(v, k)
        
        # generate pipeline flow chart
        for k, v in self.pipeline_setting.items():
            src_module, src_inst, src_group = k.split("/")
            dst_module, dst_inst, dst_tag = v.split("/")
            # get input-list
            input_list_str = "\n".join(self.wrapper_dict[src_module][src_inst]["input"])
            # get output-list
            output_list_str = "\n".join(list(self.wrapper_dict[src_module][src_inst]["map"][v].keys()))
            # get next-input-list
            if self.wrapper_dict.get(dst_module, False):
                next_input_list_str = "\n".join(self.wrapper_dict[dst_module][dst_inst]["input"])
            else:
                next_input_list_str = "null"
            # form the edge
            self.pipeline_to_save.append(
                (
                    "INPUT\n{}\nNAME\n{}/{}".format(input_list_str, src_module, src_inst),
                    "INPUT\n{}\nNAME\n{}/{}".format(next_input_list_str, dst_module, dst_inst),
                    "GROUP-{}\nTAG-{}\n{}".format(src_group, dst_tag, output_list_str),
                )
            )
        
        return self.params_dict
    
    def get_objects_dict(self, params_dict, link_config):
        self.objects_dict = {}
        condition = lambda k, v: isinstance(v, dict)
        for k in self.core_modules_order_to_create:
            if link_config.get(k, False):
                if condition(k, link_config[k]):
                    self.objects_dict[k] = utils.operate_dict_recursively(
                        src_dict=params_dict[k],
                        condition=condition,
                        operation=self._get_objects_dict_operation,
                        flag_dict=link_config[k],
                        addtion_params=k
                    )
                else:
                    self.objects_dict[k] = self._get_objects_dict_operation(params_dict[k], k)
        return self.objects_dict
    
    def load_link_config(self):
        self.link_config = utils.load_yaml(self.link_path)
        self.general_setting = self.link_config.pop(LINK_GENERAL_KEY, None)
        self.pipeline_setting = self.link_config.pop(PIPELINE_KEY, None)
        self._process_pipeline()
    
    def _process_pipeline(self):
        self.pipeline_setting = [
            [sub_item.strip() for sub_item in item.split("->")] 
            for item in self.pipeline_setting
        ]
        self.pipeline_setting = {
            item[0]: item[1]
            for item in self.pipeline_setting
        }

    def initiate_params(self):
        self.load_link_config()
        self.assert_dict = utils.load_yaml(self.assert_path)
        self.show_link()
    
    def show_link(self):
        logging.info("#####################")
        logging.info("Load link config")
        logging.info("#####################")
        for k, v in self.link_config.items():
            if k != LINK_GENERAL_KEY and k != PIPELINE_KEY:
                logging.info("... {}".format(k))
                assert isinstance(v, list)
                for item in v:
                    logging.info("... ... {}: {}".format(
                            utils.get_first_dict_key(item),
                            utils.get_first_dict_value(item)
                        )
                    )
        if self.is_confirm_first:
            _ = input("Confirm: ...")

    def create_all(self, change_dict={}):
        """
        Initialize all modules according to params dictionary.

        Args:
            change_dict (dict):
                Dictionary that overwrites certain parameters. (optional)
        
        Returns:
            dict: initialized objects dictionary.
        """
        logging.info("#####################")
        logging.info("Create objects")
        logging.info("#####################")

        # update according to argparse
        self.params_dict = self.maybe_modify_params_dict(self.params_dict, change_dict)

        # update according to general_setting in link_config
        if self.general_setting is not None:
            self.params_dict = self.maybe_modify_params_dict(
                self.params_dict, self.general_setting
            )

        self.maybe_assert_params_dict(self.params_dict)
        self.objects_dict = self.get_objects_dict(self.params_dict, self.link_config)        
        return self.objects_dict
    
    def get_certain_params_dict(self, change_list, params_dict=None):
        params_dict = (
            self.params_dict
            if params_dict is None
            else params_dict
        )
        assert isinstance(change_list, list)
        output_dict = {}
        for k in change_list:
            curr_dict = {}
            modify_list = self.convert_dict[k]
            for modify_path in modify_list:
                top_class, remain_str = modify_path.split(SPLITER)
                instance_name, attr_name = remain_str.split(ATTRSYM)
                # search the instance
                instance_dict = utils.operate_dict_recursively(
                    src_dict=params_dict[top_class],
                    condition=lambda k, v: v==instance_name,
                    operation=lambda x, params: x
                )[instance_name]
                # record parameters
                curr_dict[modify_path] = instance_dict[PARAMS_KEY][attr_name]
            output_dict[k] = curr_dict
        return output_dict

    
    def maybe_modify_params_dict(self, params_dict, change_dict={}):
        assert isinstance(change_dict, dict)
        for k, v in change_dict.items():
            modify_list = self.convert_dict[k]
            for modify_path in modify_list:
                top_class, remain_str = modify_path.split(SPLITER)
                instance_name, attr_name = remain_str.split(ATTRSYM)
                # search the instance
                instance_dict = utils.operate_dict_recursively(
                    src_dict=params_dict[top_class],
                    condition=lambda k, v: v==instance_name,
                    operation=lambda x, params: x
                ).get(instance_name, None)
                # change parameters
                if instance_dict is None:
                    logging.warn("{}/{} doesn't exist! modify failed!".format(
                        top_class, instance_name
                    ))
                    raise KeyError("Plase check key/value pairs!")
                else:
                    instance_dict[PARAMS_KEY][attr_name] = v
                    logging.info("{}/{}/{} is changed to {}".format(
                        top_class, instance_name, attr_name, v
                    ))

        return params_dict

    def maybe_assert_params_dict(self, params_dict):
        logging.warning("'maybe_assert_params_dict' is not implemented!")



import torch
import torch.nn.functional as F

from ..base_collector import BaseCollector

class DAMLCollector(BaseCollector):
    """
    NOTE: only support Triplet-Loss.

    Paper: `Deep Adversarial Metric Learning <https://openaccess.thecvf.com/content_cvpr_2018/html/Duan_Deep_Adversarial_Metric_CVPR_2018_paper.html>`_

    Training steps:

    1. pretrain the deep metric learning model without the hard negative generator;
    2. initialize the generator adversarial to the pre-trained metric;
    3. jointly optimize both networks during each iteration end-to-end

    Three losses for hard negative generation:

    1. the synthetic samples should be close to the anchor in the original feature space;
    2. the synthetic samples should perserve the annotation information;
    3. the synthetic samples should be misclassified by the learned metric

    Default backbone structure:

    1. trunk: ``GoogLeNet``
    2. embedder: one-layer perceptron
    3. generator: three-layer perceptron

    Args:
        embedder (torch.nn.Module): embedder model (default: one-layer perceptron)
        generator (torch.nn.Module): generator model (default: three-layer perceptron)
        lambda_0 (int): default: 1
        lambda_1 (int): default: 1
        lambda_2 (int): default: 50
        alpha (int): default: 1
    """
    def __init__(
        self,
        embedder,
        generator,
        lambda_0=1,
        lambda_1=1,
        lambda_2=50,
        alpha=1,
        *args,
        **kwargs
    ):
        super(DAMLCollector, self).__init__(*args, **kwargs)
        self.embedder = embedder
        self.generator = generator
        self.lambda_0 = lambda_0
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.alpha = alpha

    def update(self, trainer):
        pass

    def forward(
        self,
        data,
        embeddings,
        features,
        labels
    ) -> tuple:
        """
        There are four losses to be computed in ``collect`` function (All losses will be computed in this function, i.e. NOT pass to ``selectors`` or ``losses`` modules)

        :math:`loss_{total} = \lambda_0 \\times loss_m + \lambda_1 \\times loss_{reg} + \lambda_2 \\times loss_{adv} + loss_{hard}`

        :math:`loss_m = mean(ReLU(D_{ap emb} - D_{an emb} - \\alpha))`

        :math:`loss_{adv} = mean(ReLU(D_{an feat} - D_{ap feat} - \\alpha))`

        :math:`loss_{reg} = mean(|f_{syn} - f_{neg}|^2_2)`

        :math:`loss_{hard} = mean(|f_{syn} - f_{anchor}|^2_2)`
        """
        # get a triplet
        matches = (labels.unsqueeze(1) == labels.unsqueeze(0)).byte()
        diffs = matches ^ 1
        matches.fill_diagonal_(0)
        triplets = matches.unsqueeze(2) * diffs.unsqueeze(1)
        anc_idx, pos_idx, neg_idx = torch.where(triplets)

        anc_features = features[anc_idx]
        pos_features = features[pos_idx]
        neg_features = features[neg_idx]
        anc_embeddings = embeddings[anc_idx]
        pos_embedding = embeddings[pos_idx]

        # generate synthetic samples
        cat_features = torch.cat([anc_features, pos_features, neg_features], dim=-1)
        syn_features = self.generator(cat_features)
        syn_embeddings = self.embedder(syn_features)

        # compute loss_reg
        loss_reg = torch.mean(
            torch.sum((syn_features - neg_features)**2, dim=-1)
        )

        # compute loss_hard
        loss_hard = torch.mean(
            torch.sum((syn_features - anc_features)**2, dim=-1)
        )

        # compute loss_adv
        dist_an_feat = torch.sum(((anc_features - syn_features)**2), dim=-1)
        dist_ap_feat = torch.sum(((anc_features - pos_features)**2), dim=-1)
        loss_adv = torch.mean(
            F.relu(dist_an_feat - dist_ap_feat - self.alpha)
        )

        # compute loss_m
        dist_an_emb = torch.sum(((anc_embeddings - syn_embeddings)**2), dim=-1)
        dist_ap_emb = torch.sum(((anc_embeddings - pos_embedding)**2), dim=-1) 
        loss_m = torch.mean(
            F.relu(dist_ap_emb - dist_an_emb - self.alpha)
        )

        return (
            loss_m,
            self.lambda_0,
            loss_reg,
            self.lambda_1,
            loss_hard,
            1,
            loss_adv,
            self.lambda_2
        )

    


import torch

from .large_margin_softmax_loss import LargeMarginSoftmaxLoss

class CosFaceLoss(LargeMarginSoftmaxLoss):
    """
    paper: `CosFace: Large Margin Cosine Loss for Deep Face Recognition <http://openaccess.thecvf.com/content_cvpr_2018/html/Wang_CosFace_Large_Margin_CVPR_2018_paper.html>`_
    """
    def __init__(self, margin=0.35, scale=64, **kwargs):
        super(CosFaceLoss, self).__init__(margin=margin, scale=scale, **kwargs)
    
    def initiate_margin(self):
        pass

    def modify_cosine_of_target_classes(self, cosine_of_target_classes):
        return cosine_of_target_classes - self.margin
    
    


# recorder protocol
TO_RECORD_LIST = 'to_record_list'
CSV_FOLDER = 'csv'
MODEL_FOLDER = 'model'
BOARD_FOLDER = 'board'
WANDB_FOLDER = 'wandb'
TO_SAVE_LIST = 'to_save_list'
DEVICE = 'device'
BEST_MODEL_SUFFIX = '_best.ckpt'
STEP_MODEL_SUFFIX = '_{}.ckpt'
LOAD_EXCEPT_FUNC = lambda x: x[7:]


import torch
import torch.distributed as dist
import logging
from tqdm import tqdm

from ..misc import utils, Storage
from .loss_handler import LossHandler
from ...config.setting.recorder_setting import (
    TO_RECORD_LIST,
    TO_SAVE_LIST
)

class BaseTrainer:
    """
    ``BaseTrainer`` takes charge of training.

    ``storage`` holds all the intermediate variables such as ``data``, ``embeddings``, etc. All other modules, such as ``collectors``, ``selectors``, etc, access thier own parameter list to get the corresponding parameters.

    ``loss_handler`` will compute weighted loss.

    Args:
        batch_size (int):
            Batch size.
        freeze_trunk_batchnorm (bool):
            Whether to freeze batch normalization layer.
        dataset_num_workers (int):
            Number of process to load data.
        is_distributed (bool):
            Whether to use distibuted mode.
    """
    def __init__(
        self,
        batch_size,
        wrapper_params,
        freeze_trunk_batchnorm=False,
        dataset_num_workers=8,
        is_distributed=False,
    ):
        self.batch_size = batch_size
        self.freeze_trunk_batchnorm = freeze_trunk_batchnorm
        self.dataset_num_workers = dataset_num_workers
        self.is_distributed = is_distributed
        
        self.epochs = 0
        self.storage = Storage(wrapper_params)
        self.initiate_property()
    
    """
    Initialization
    """
    
    def initiate_property(self):
        # to save list (for recorders)
        setattr(
            self,
            TO_SAVE_LIST,
            ["models", "collectors", "losses"]
        )
        # recordable list
        self.recordable_object_list = ["models"]
        # trainable list
        self.trainable_object_list = [
            "models",
            "collectors"
        ]
        
    @property
    def models_forward_order(self):
        return ["trunk", "embedder"]
    
    """
    Set and Get
    """

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size

    def set_distributed(self, flag=True):
        self.is_distributed = flag
    
    def set_activated_optims(self, optim_list=None):
        self.optim_list = (
            list(self.optimizers.keys()) 
            if optim_list is None
            else optim_list
        )
        logging.info("Set activated optims: {}".format(self.optim_list))
        
    """
    Train
    """

    def prepare(
        self,
        collectors,
        selectors,
        losses,
        models,
        optimizers,
        datasets,
        schedulers=None,
        gradclipper=None,
        samplers=None,
        collatefns=None,
        device=None,
        recorders=None
    ):
        """
        Load modules to prepare training.
        """
        # pass parameters
        self.collectors = collectors
        self.selectors = selectors
        self.losses = losses
        self.models = models
        self.optimizers = optimizers
        self.datasets = datasets
        self.schedulers = schedulers
        self.gradclipper = gradclipper
        self.samplers = samplers
        self.collatefns = collatefns
        self.device = device
        self.loss_handler = LossHandler()
        self.recorders = utils.get_default(recorders, "recorders")

        self.optim_list = list(self.optimizers.keys())


    def train(self, epochs=None):
        """
        Start training.

        Args:
            epochs (int):
                Epoch to start.
        """
        # start to train
        if epochs is not None:
            self.epochs = epochs
        self.initiate_dataloader()
        logging.info("TRAIN EPOCH: {}".format(self.epochs))
        self.pipeline()
        self.epochs += 1
        # self.release_memory()
    
    def pipeline(self):
        # use global collector 
        self.use_global_collector()
        self.set_to_train()
        self.show_lr()

        # start training
        self.pbar = tqdm(range(self.iterations_per_epoch))
        for self.iteration in self.pbar:
            self.prepare_forward()
            self.forward_pipeline()
            self.backward_and_update()
            self.update_record(self.recorders)
            self.pbar.set_description(
                "Loss={:.4f}".format(
                    self.loss_handler.get_total(),
                )
            )
        self.loss_handler.average_losses()
        self.step_schedulers(metrics=self.loss_handler.get_total())
    
    def show_lr(self):
        default_idx = 0
        for k, v in self.optimizers.items():
            lr = v.param_groups[default_idx]["lr"]
            logging.info("{} optimizer's lr: {}".format(k, lr))
    
    def forward_pipeline(self):
        self.forward_models()
        self.forward_collectors()
        self.forward_selectors()
        self.forward_losses()
    
    def initiate_dataloader(self):
        logging.info(
            "{}: Initiating dataloader".format(
                self.__class__.__name__
            )
        )
        
        # more initialization
        self.sampler, self.collate_fn = None, None
        self._initiate_dataloader_sampler_collatefn()

        # get dataloader
        self.dataloader = torch.utils.data.DataLoader(
            dataset=self.datasets["train"],
            batch_size=int(self.batch_size),
            sampler=self.sampler,
            drop_last=True,
            pin_memory=False,
            shuffle=self.samplers is None,
            num_workers=self.dataset_num_workers,
            collate_fn=self.collate_fn
        )
        self.iterations_per_epoch = len(self.dataloader)
        self.dataloader_iter = iter(self.dataloader)
    
    def _initiate_dataloader_sampler_collatefn(self):
        # extract the sampler
        if self.is_distributed:
            if self.samplers is not None:
                logging.warn("class samplers can't be used in distribued training mode!")
            self.samplers = {}
            self.samplers["train"] = torch.utils.data.distributed.DistributedSampler(
                dataset=self.datasets["train"],
                shuffle=True,
            )
            self.sampler = self.samplers["train"]
            logging.info("Get distributed sampler {}".format(
                self.sampler.__class__.__name__,
            ))
        else:
            self.sampler = (
                self.samplers["train"] 
                if self.samplers is not None
                else None
            )
        
        # extract the collate_fn
        self.collate_fn = (
            self.collatefns["train"]
            if self.collatefns is not None
            else None
        )

    def set_to_train(self):
        for trainable_name in self.trainable_object_list:
            trainable_object = getattr(self, trainable_name, None)
            if trainable_object is None:
                logging.warn(
                    "{} is not a member of trainer".format(
                        trainable_name
                    )
                )
            else:
                for v in trainable_object.values():
                    v.train()
        # maybe some parameters should be frozen
        self._maybe_freeze_batchnorm()
    
    def _maybe_freeze_batchnorm(self):
        if self.freeze_trunk_batchnorm:
            self.models["trunk"].apply(
                utils.set_layers_to_eval("BatchNorm")
            )
        
    def use_global_collector(self):
        for collector in self.collectors.values():
            if collector.is_global_collector:
                logging.info("Global collector updating...")
                collector.global_update(self)
                self.release_memory()
    
    def prepare_forward(self):
        # set sampler
        self._prepare_forward_set_sampler()

        # zero loss-values
        self.loss_handler.zero_losses()

        # zero grad
        for v in self.models.values():
            v.zero_grad()
        for v in self.optimizers.values():
            v.zero_grad()

        # get batch
        self._prepare_forward_get_batch()
    
    def _prepare_forward_set_sampler(self):
        if self.is_distributed:
            self.sampler.set_epoch(self.epochs)
    
    @property
    def info_dict_to_device(self):
        return [
            "data", 
            "labels"
        ]
    
    def _prepare_forward_get_batch(self):
        info_dict = next(self.dataloader_iter)
        for key in info_dict.keys():
            setattr(self.storage, key, info_dict[key])
        self.storage.tensors_to_device(self.info_dict_to_device, self.device)

    def forward_models(self):
        # get data and labels
        data = self.storage.get("data")
        labels = self.storage.get("labels")

        # forward backbone (trunk model)
        features = self.models["trunk"](data)

        # if distributed
        if self.is_distributed:
            features, labels = utils.distributed_gather_objects(
                features, labels
            )

        self.storage.features = features
        self.storage.labels = labels
        self.storage.indices_dict["models"] = {"embedder": {"": {}}}

        self.storage.update(self.models["embedder"], cur_module="models")

    def forward_collectors(self):
        # update collector
        self.update_collectors()
        # forward collector
        self.storage.update(self.collectors, cur_module="collectors")
    
    def update_collectors(self):
        for collector in self.collectors.values():
            if isinstance(collector, (torch.nn.DataParallel, torch.nn.parallel.DistributedDataParallel)):
                collector.module.update(self)
            else:
                collector.update(self)
    
    def forward_selectors(self):
        self.storage.update(self.selectors, cur_module="selectors")

    def forward_losses(self):
        value_dict = self.storage.update(self.losses, cur_module="losses")

        # update loss_values
        self.loss_handler.update_losses(self.storage.return_loss_dict())

    def backward_and_update(self):
        self._backward_and_update_preprocess()

        self.loss_handler.backward()
        # clip gradients
        if self.gradclipper is not None:
            for v in self.gradclipper.values():
                v()
        # step optimizers
        for k in self.optim_list:
            self.optimizers[k].step()
    
    def _backward_and_update_preprocess(self):
        # for distributed
        if self.is_distributed:
            # dist.barrier()
            dist.all_reduce(
                tensor=self.loss_values["total_loss"],
                op=dist.ReduceOp.SUM
            )

    def step_schedulers(self, **kwargs):
        """
        All schedulers step at the end of each epoch for the moment
        """
        if self.schedulers is not None:
            for k in self.optim_list:
                v = self.schedulers[k]
                if isinstance(v, torch.optim.lr_scheduler.ReduceLROnPlateau):
                    v.step(**kwargs)
                else:
                    v.step()

    def update_record(self, recorders=None):
        if recorders is not None:
            total_iterations = (
                self.epochs * self.iterations_per_epoch 
                + self.iteration
            )
            # update loss
            self.loss_handler.record_losses(recorders, total_iterations)

            # update other statistics
            for recordable_name in self.recordable_object_list:
                recordable_object = getattr(self, recordable_name)
                for k, v in recordable_object.items():
                    if k == "trunk": # TODO:
                        continue
                    to_record_obj = (
                        v.module
                        if isinstance(v, torch.nn.DataParallel) or isinstance(v, torch.nn.parallel.DistributedDataParallel)
                        else v
                    )
                    data, _ = recorders.get_data(to_record_obj, k)
                    recorders.update(data, total_iterations)

    def release_memory(self):
        torch.cuda.empty_cache()


from .factories.metrics_creator import metricsCreator
from .factories.collectors_creator import collectorsCreator
from .factories.selectors_creator import selectorsCreator
from .factories.losses_creator import lossesCreator
from .factories.models_creator import modelsCreator
from .factories.evaluators_creator import evaluatorsCreator
from .factories.optimizers_creator import optimizersCreator
from .factories.schedulers_creator import schedulersCreator
from .factories.grad_clipper_creator import gradclipperCreator
from .factories.transforms_creator import transformsCreator
from .factories.datasets_creator import datasetsCreator
from .factories.samplers_creator import samplersCreator
from .factories.recorders_creator import recordersCreator
from .factories.managers_creator import managersCreator
from .factories.trainers_creator import trainersCreator
from .factories.testers_creator import testersCreator
from .factories.collatefns_creator import collatefnsCreator

class CreatorManager:
    def __init__(self):
        self.registered_creators = [
            "metricsCreator",
            "collectorsCreator",
            "selectorsCreator",
            "lossesCreator",
            "modelsCreator",
            "evaluatorsCreator",
            "optimizersCreator",
            "schedulersCreator",
            "gradclipperCreator",
            "transformsCreator",
            "datasetsCreator",
            "samplersCreator",
            "recordersCreator",
            "managersCreator",
            "trainersCreator",
            "testersCreator",
            "collatefnsCreator"
        ]
        self.creators_dict = {
            name: eval(name)()
            for name in self.registered_creators
        }
    
    def register_packages(self, module_name, extra_package):
        """
        Register new packages into the specific module-creator.

        Args:
            module_name (str):
                The specific module-creator.
            extra_package (list or module):
                Extra packages to be added.
        """
        creator_name = module_name + 'Creator'
        creator_item = self.creators_dict[creator_name]
        creator_item.register_packages(extra_package)

    def create(self, module_type, module_params):
        creator_name = module_type + 'Creator'
        creator_item = self.creators_dict[creator_name]
        creator_output = creator_item(module_params)
        return creator_output


import torch
import math
from copy import deepcopy
import logging
import torch.distributed as dist 
from ..base_collector import BaseCollector
from ...misc import utils

class MoCoCollector(BaseCollector):
    """
    Paper: `Momentum Contrast for Unsupervised Visual Representation Learning <https://openaccess.thecvf.com/content_CVPR_2020/html/He_Momentum_Contrast_for_Unsupervised_Visual_Representation_Learning_CVPR_2020_paper.html>`_

    Use Momentum Contrast (MoCo) for unsupervised visual representation learning. This code is modified from: https://github.com/facebookresearch/moco. In this paper, a dynamic dictionary with a queue and a moving-averaged encoder are built.

    Args:
        query_trunk (torch.nn.Module): default: ResNet50
        query_embedder (torch.nn.Module): multi-layer perceptron
        embeddings_dim (int): dimension of embeddings. default: 128
        bank_size (int): size of the memory bank. default: 65536
        m (float): weight of moving-average. default: 0.999
        T (float): coefficient of softmax
    """
    def __init__(
        self,
        query_trunk,
        query_embedder,
        embeddings_dim=128,
        bank_size=65536,
        m=0.999,
        T=0.07,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs) 
        self.embeddings_dim = embeddings_dim
        self.bank_size = bank_size
        self.m = m
        self.T = T
        self.key_trunk = deepcopy(query_trunk)
        self.key_embedder = deepcopy(query_embedder)
        self.initiate_params()
    
    def initiate_params(self):
        """
        Cancel the gradient of key_trunk and key_embedder
        """
        for param_k in self.key_trunk.parameters():
            param_k.requires_grad = False # not update by gradient
        for param_ke in self.key_embedder.parameters():
            param_ke.requires_grad = False # not update by gradient

        self.register_buffer("queue", torch.randn(self.bank_size, self.embeddings_dim))
        self.register_buffer("queue_ptr", torch.zeros(1, dtype=torch.long))
    
    @torch.no_grad()
    def _momentum_update_key_encoder(self, models_dict: dict):
        """
        Momentum update of the key model (encoder)
        """
        for param_q, param_k in zip(models_dict["trunk"].parameters(), self.key_trunk.parameters()):
            param_k.data = param_k.data * self.m + param_q.data * (1. - self.m)
        for param_qe, param_ke in zip(models_dict["embedder"].parameters(), self.key_embedder.parameters()):
            param_ke.data = param_ke.data * self.m + param_qe.data * (1. - self.m)
    
    @torch.no_grad()
    def _dequeue_and_enqueue(self, keys: torch.Tensor):
        # gather keys before updating queue
        if dist.is_initialized():
            keys, = utils.distributed_gather_objects(keys)
        batch_size = keys.shape[0]

        # get pointer
        ptr = int(self.queue_ptr)
        assert self.bank_size % batch_size == 0 # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.queue[ptr:(ptr+batch_size), :] = keys
        ptr = (ptr + batch_size) % self.bank_size
        self.queue_ptr[0] = ptr
    
    @torch.no_grad()
    def _batch_shuffle_ddp(self, x: torch.Tensor):
        """
        Batch shuffle, for making use of BatchNorm.

        Note:
            Only support DistributedDataParallel (DDP) model.
        """
        # gather from all gpus
        device = x.device
        batch_size_this = x.shape[0]
        x_gather, = utils.distributed_gather_objects(x)
        batch_size_all = x_gather.shape[0]
        num_gpus = batch_size_all // batch_size_this

        # random shuffle index
        idx_shuffle = torch.randperm(batch_size_all).to(device)

        # broadcast to all gpus
        dist.broadcast(idx_shuffle, src=0)

        # index for restoring
        idx_unshuffle = torch.argsort(idx_shuffle)

        # shuffled index for this gpu
        gpu_idx = dist.get_rank()
        idx_this = idx_shuffle.view(num_gpus, -1)[gpu_idx]

        return x_gather[idx_this], idx_unshuffle
    
    @torch.no_grad()
    def _batch_unshuffle_ddp(self, x: torch.Tensor, idx_unshuffle: torch.Tensor):
        """
        Undo batch shuffle.

        Note:
            Only support DistributedDataParallel (DDP) model.
        """
        # gather from all gpus
        batch_size_this = x.shape[0]
        x_gather, = utils.distributed_gather_objects(x)
        batch_size_all = x_gather.shape[0]
        num_gpus = batch_size_all // batch_size_this

        # restored index for this gpu
        gpu_idx = dist.get_rank()
        idx_this = idx_unshuffle.view(num_gpus, -1)[gpu_idx]

        return x_gather[idx_this]
    
    @torch.no_grad()
    def update(self, trainer):
        """
        Update key encoder using moving-average.
        """
        with torch.no_grad():
            self._momentum_update_key_encoder(
                trainer.models
            )
    
    def forward(self, data, embeddings) -> tuple:
        """
        Maintain a large memory bank to boost unsupervised learning performance.

        Args:
            data (torch.Tensor):
                A batch of key images. size: :math:`B \\times C \\times H \\times W`
            embeddings (torch.Tensor):
                A batch of query embeddings. size: :math:`B \\times dim`
        """
        # compute key features
        device = embeddings.device
        batch_size = embeddings.shape[0]
        if dist.is_initialized():
            world_size = dist.get_world_size()
            batch_size = batch_size // world_size
            sub_embeddings = embeddings.view(dist.get_world_size(), batch_size, -1)[dist.get_rank()]
        else:
            sub_embeddings = embeddings
        
        with torch.no_grad(): # no gradient to keys
            # shuffle for making use of BN
            if dist.is_initialized():
                data, idx_unshuffle = self._batch_shuffle_ddp(data)
            
            # get key embeddings
            keys = self.get_key_embeddings(data) # keys: N x C

            # undo shuffle
            if dist.is_initialized():
                keys = self._batch_unshuffle_ddp(keys, idx_unshuffle)
        
        # compute matrix
        metric_mat = self.metric(
            sub_embeddings,
            keys.clone().detach(),
            self.queue.clone().detach()
        )

        # apply temperature
        metric_mat /= self.T 

        # generate labels 
        # TODO: 为什么不用管queue中的正样本？
        row_labels = torch.arange(batch_size).to(device).unsqueeze(-1)
        neg_labels = torch.arange(batch_size, self.bank_size + batch_size).unsqueeze(0).repeat(batch_size, 1).to(device)
        col_labels = torch.cat(
            [
                row_labels,
                neg_labels
            ],
            dim=-1
        )

        # dequeue and enqueue
        self._dequeue_and_enqueue(keys)

        is_same_source = False
        return (
            metric_mat,
            row_labels,
            col_labels,
            is_same_source
        )
    
    def get_key_embeddings(self, data):
        trunk_output = self.key_trunk(data)
        embeddings = self.key_embedder(trunk_output)
        return embeddings


import yaml
import inspect
from copy import deepcopy

import torch
import torch.distributed as dist 
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.nn import DataParallel as DP 

from ...config.setting.launcher_setting import (
    DEFAULT_NAME
)

class gradclipper:
    def __init__(self, parameters, max_norm):
        self.parameters = parameters
        self.max_norm = max_norm

    def __call__(self):
        torch.nn.utils.clip_grad_norm_(self.parameters, self.max_norm)

def operate_dict_recursively(src_dict, condition, operation, flag_dict=None, addtion_params=None):
    flag_dict = (
        flag_dict
        if flag_dict is not None
        else src_dict
    )
    dst_dict = dict()
    for k, v in flag_dict.items():
        if condition(k, v):
            dst_dict[k] = operate_dict_recursively(
                src_dict=src_dict[k],
                condition=condition,
                operation=operation,
                flag_dict=flag_dict[k],
                addtion_params=addtion_params
            )
        else:
            dst_dict[k] = operation(v, addtion_params)
    return dst_dict

def load_yaml(yaml_path):
    return yaml.load(
        stream=open(yaml_path, mode='r'),
        Loader=yaml.FullLoader
    )

def get_first_dict_key(data_dict):
    return list(data_dict.keys())[0]

def get_first_dict_value(data_dict):
    return list(data_dict.values())[0]

def get_last_linear(model):
    for name in ['fc', 'last_linear']:
        last_layer = getattr(model, name, None)
        if last_layer is not None:
            return last_layer, name
    raise KeyError("Structure missing: There is no fc or last_linear in the givin model!")

def set_last_linear(model, target):
    setattr(model, get_last_linear(model)[1], target) 

def set_layers_to_eval(layer_name):
    def set_to_eval(m):
        classname = m.__class__.__name__
        if classname.find(layer_name) != -1:
            m.eval()
    return set_to_eval

def func_params_mediator(params_bank, init_func):
    """
    Pass parameters to a function according to func's parameters list (through inspect package)

    Args:
        params_bank (list):
            List of the parameters bank. Each item must be the instance of ``dictionary`` or ``object``.
        init_func (func):
            Function to be called.
    
    Returns:
        The output of the function.
    """
    # get the completed params
    assert isinstance(params_bank, list)

    # check
    signature = inspect.getfullargspec(init_func)
    all_params = signature.args
    params_len = len(all_params)
    optional_len = (
        len(signature.defaults) 
        if signature.defaults is not None
        else 0
    )
    required_params = [item for item in all_params[:(params_len-optional_len)] if item != 'self']

    # construct params_dict
    curr_params = {}
    for params_name in all_params:
        for curr_bank in params_bank:
            if isinstance(curr_bank, dict):
                params_value = curr_bank.get(params_name, None)
            else:
                params_value = getattr(curr_bank, params_name, None)
            
            if params_value is not None:
                curr_params[params_name] = params_value

    assert all([item in list(curr_params.keys()) for item in required_params])
    return init_func(**curr_params)

def get_default(input_object, module_name):
    output = None
    if isinstance(input_object, dict):
        output = input_object.get(module_name, None)
        if output is None:
            output = input_object.get(DEFAULT_NAME, None)
        else:
            output = output.get(DEFAULT_NAME, None)
        return output
    else:
        return input_object

def dict_to_command(input_dict, ):
    command = ""
    for k, v in input_dict.items():
        command += "--{} {} ".format(
            k, v
        )
    return command

"""
Distributed utilities
"""
def distributed_gather_objects(*objects_list, rank=None, world_size=None):
    gathered_objects_list = []
    world_size = (
        dist.get_world_size()
        if world_size is None
        else world_size
    )
    rank = (
        dist.get_rank()
        if rank is None
        else rank
    )
    for objects_item in list(objects_list):
        curr_gather_list = [
            torch.zeros_like(objects_item)
            for _ in range(world_size)
        ]
        dist.all_gather(
            tensor_list=curr_gather_list,
            tensor=objects_item.contiguous()
        )
        curr_gather_list[rank] = objects_item
        curr_gather = torch.cat(curr_gather_list)
        gathered_objects_list.append(curr_gather)
    return tuple(gathered_objects_list)



import torch
import torch.nn as nn 

from ..modules import WithRecorder

"""
Normal multi-layer perceptron
"""

class MLP(WithRecorder):
    '''
    Normal multi-layer perceptron.

    Args:
        layer_size_list (list):
            The numbers of neurals in each layer.
        first_relu (bool):
            Whether to set ReLU at the beginning of the MLP.
        last_relu (bool):
            Whether to set ReLU at the end of the MLP.
    
    Example:
        >>> model = MLP(layer_size_list=[512, 100], first_relu=False)
    '''
    def __init__(self, layer_size_list, first_relu=True, last_relu=False, output_dim=None):
        super(MLP, self).__init__()
        # check data type
        if output_dim is not None:
            layer_size_list[-1] = int(output_dim)
        self.layer_size_list = [int(item) for item in layer_size_list]
        self.first_relu = first_relu
        self.last_relu = last_relu
        # construct MLP
        layer_list = [nn.ReLU(inplace=True)] if first_relu else []
        for idx in range(len(layer_size_list)-1):
            layer_list.append(nn.Linear(layer_size_list[idx], layer_size_list[idx+1]))
            if idx != len(layer_size_list) - 2:
                layer_list.append(nn.ReLU(inplace=True))
        if last_relu:
            layer_list.append(nn.ReLU(inplace=True))
        # compose 
        self.net = nn.Sequential(*layer_list)
        self.last_linear = self.net[-1]
    
    def forward(self, features):
        return self.net(features)

"""
multi-layer perceptron with batch normalization
"""

class BatchNormLayer(WithRecorder):
    def __init__(self, in_dim, out_dim, is_bn=True, is_relu=True, **kwargs):
        super(BatchNormLayer, self).__init__(**kwargs)
        layers_list = []
        # add linear 
        layers_list.append(nn.Linear(in_dim, out_dim))
        # add batch norm
        if is_bn:
            layers_list.append(nn.BatchNorm1d(out_dim))
        # add ReLU
        if is_relu:
            layers_list.append(nn.ReLU(inplace=True))
        # compose
        self.net = nn.Sequential(*layers_list)
        
    def forward(self, x):
        return self.net(x)

class BatchNormMLP(WithRecorder):
    """
    Multi-layer perceptron with batch normalization.

    Args:
        layer_size_list (list):     
            The numbers of neurals in each layer. (N + 1)
        relu_list (list):
            Whether relu is added.
        bn_list (list):
            Whether bn is added.
        first_bn (bool):
            Whether to set BN at the beginning of the MLP.
    
    Example:
        >>> model = BatchNormMLP(
            layer_size_list=[512, 512, 1024],
            relu_list=[True, False],
            bn_list=[True, False],
            first_bn=False
        )
    """
    def __init__(self, layer_size_list, relu_list, bn_list, first_bn=False, **kwargs):
        super(BatchNormMLP, self).__init__(**kwargs)
        layer_size_list = [int(item) for item in layer_size_list]
        self.layer_size_list = layer_size_list
        self.relu_list = relu_list
        self.bn_list = bn_list
        self.layers_num = len(self.relu_list)
        self.first_bn = first_bn
        
        layers_list = []
        if self.first_bn:
            layers_list.append(
                nn.BatchNorm1d(self.layer_size_list[0])
            )

        for i in range(self.layers_num):
            layers_list.append(
                BatchNormLayer(
                    in_dim=self.layer_size_list[i],
                    out_dim=self.layer_size_list[i+1],
                    is_bn=self.bn_list[i],
                    is_relu=self.relu_list[i]
                )
            )
        self.net = nn.Sequential(*layers_list)
    
    def forward(self, features):
        return self.net(features)

# if __name__ == "__main__":
#     model = BatchNormMLP(
#         layer_size_list=[512, 512, 1024],
#         relu_list=[True, False],
#         bn_list=[True, False],
#         first_bn=False
#     )
#     print(model)

if __name__ == '__main__':
    model = MLP(layer_size_list=[512, 100], first_relu=False)
    print(model)


import torch
import logging
from .base_creator import BaseCreator
from ....core import models
from ...misc import utils

class modelsCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [models]
    
    def maybe_modify_object(self, module_object: torch.nn.Module):
        if self.creator_mode is None:
            pass
        elif self.creator_mode == "delete_last_linear":
            if not isinstance(module_object, models.MLP):
                utils.set_last_linear(module_object, models.Identity())
            logging.info("Creator MODE: Delete last linear: {}".format(module_object.__class__.__name__))
        elif self.creator_mode == "freeze_all":
            if not isinstance(module_object, models.MLP):
                utils.set_last_linear(module_object, models.Identity())
            for param in module_object.parameters():
                param.requires_grad = False
            logging.info("Creator MODE: freeze all: {}".format(module_object.__class__.__name__))
        elif self.creator_mode == "freeze_except_last_linear":
            if not isinstance(module_object, models.MLP):
                utils.set_last_linear(module_object, models.Identity())
            for name, param in module_object.named_parameters():
                if name not in []:
                    param.requires_grad = False
            logging.info("Creator MODE: freeze except last linear: {}".format(module_object.__class__.__name__))
        else:
            raise KeyError("Invalid creator_type: {}, in {}".format(
                self.creator_mode, self.__class__.__name__
            ))
        return module_object


"""
Note: 
    **Tmux** software support is required.

This module is designed for assisting distributed trainning and hyperparameter searching. There are two situations where this module can be used:

1. **Distributed training**. PyTorch distributed API need multi-device support. Tmux software is used for launching multi processes.

2. **Hyperparameter searching**. Tmux software is used for launching each experiment which corresponds to a certain hyperparameter combination.

More information about `Tmux <https://github.com/tmux/tmux>`_.
"""

from .manager import (
    TmuxManager,
    clear_tmux
)


import torch
import numpy as np 

from .large_margin_softmax_loss import LargeMarginSoftmaxLoss

class ArcFaceLoss(LargeMarginSoftmaxLoss):
    """
    paper: `ArcFace: Additive Angular Margin Loss for Deep Face Recognition <https://openaccess.thecvf.com/content_CVPR_2019/html/Deng_ArcFace_Additive_Angular_Margin_Loss_for_Deep_Face_Recognition_CVPR_2019_paper.html>`_
    """
    def __init__(self, margin=28.6, scale=64, **kwargs):
        super(ArcFaceLoss, self).__init__(margin=margin, scale=scale, **kwargs)

    def initiate_margin(self):
        self.margin = np.radians(self.margin)
    
    def modify_cosine_of_target_classes(self, cosine_of_target_classes):
        angles = self.get_angles(cosine_of_target_classes)
        return torch.cos(angles + self.margin)


import torchvision.transforms.functional as F
from PIL import Image


class ConvertToBGR(object):
    """
    Converts a PIL image from RGB to BGR.
    """
    def __init__(self):
        pass

    def __call__(self, img):
        if len(img.getbands()) == 3:
            r, g, b = img.split()
        else:
            r, g, b = img, img, img
        img = Image.merge("RGB", (b, g, r))
        return img

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)


class Multiplier(object):
    """
    Multiply the pixel value by a constant.
    """
    def __init__(self, multiple):
        self.multiple = multiple

    def __call__(self, img):
        return img*self.multiple

    def __repr__(self):
        return "{}(multiple={})".format(self.__class__.__name__, self.multiple)


import csv
import argparse
import os

class ParserWithConvert:
    """
    We can employ this class to parser and convert args, which should work with ConfigHandler.

    Given the path of a .csv file which records the descriptions of args, 
    this class will output a `Parser` instance and convert dictionary.

    Args:
        csv_path (str):
            Path to the .csv file.
        name (str, optional):
            Name of the project.
    """
    def __init__(
        self,
        csv_path,
        name="",
    ):
        self.csv_path = csv_path
        self.name = name
        self.prefix = "config_"

        self.setting = None
    
    def render(self):
        """
        Generate parsed args and convert dictionary.

        Return:
            tuple. (args, convert)
        """
        self.read_csv()
        opt = self.get_parser()
        convert = self.get_convert()
        return opt, convert
    
    def read_csv(self):
        self.info_dict = {}
        with open(self.csv_path, mode="r", encoding="utf-8") as f:
            csv_f = csv.reader(f)
            headers = next(csv_f)
            headers = [item.strip() for item in headers]
            headers = [
                item if idx < 4 else self.prefix + item 
                for idx, item in enumerate(headers)
            ]
            for idx, row in enumerate(csv_f):
                row = [
                    eval(item) if item else ""
                    for item in row
                ]
                self.info_dict[idx] = dict(zip(headers, row))

    def get_parser(self):
        parser = argparse.ArgumentParser(self.name)
        for idx, values in self.info_dict.items():
            if values["name"] == "--setting":
                raise KeyError("<--setting> is duplicated in the command line arguments!")
            
            # for action args
            if isinstance(values["type"], str) and "store" in values["type"]:
                parser.add_argument(
                    values["name"],
                    action=values["type"],
                    default=values["default"],
                    help=values["help"]
                )
            # for list args
            elif isinstance(values["default"], list):
                parser.add_argument(
                    values["name"],
                    type=values["type"],
                    default=values["default"],
                    help=values["help"],
                    nargs="+",
                )
            # for other args
            else:
                parser.add_argument(
                    values["name"],
                    type=values["type"],
                    default=values["default"],
                    help=values["help"]
                )
        # add <setting> argument
        parser.add_argument(
            "--setting",
            type=str,
            default="default"
        )
        # get the args
        opt = parser.parse_args()
        self.setting = opt.setting
        return opt

    def get_convert(self):
        convert = {}
        for idx, values in self.info_dict.items():
            if values[self.prefix + self.setting]:
                convert[values["name"].strip('-')] = values[self.prefix + self.setting]
        return convert

if __name__ == "__main__":
    import os
    workspace = os.environ["WORKSPACE"]
    path = os.path.join(
        workspace,
        "code/GeDML/test.csv"
    )
    testParser = ParserWithConvert(csv_path=path)
    opt, convert = testParser.render()
    pass


import torch
from collections import defaultdict

class LossHandler:
    def __init__(self):
        self.weight_values = {}
        self.loss_values = {"total": torch.Tensor(0)}
        self.avg_loss_list = defaultdict(list)
        self.avg_loss_values = {}

    def get_total(self, is_avg_value=False):
        if is_avg_value:
            return self.avg_loss_values["total"]
        else:
            return self.loss_values["total"] 
    
    def backward(self):
        self.loss_values["total"].backward()
    
    def zero_losses(self):
        for k in self.loss_values.keys():
            self.loss_values[k] = 0
    
    def record_losses(self, recorder, total_iterations):
        recorder.update(
            {"loss_values": self.loss_values},
            total_iterations
        )
        recorder.update(
            {"loss_weight": self.weight_values}, 
            total_iterations
        )
        
    def update_losses(self, loss_dict: dict):
        """
        Format of loss_dict:
        {
            "default(loss_module_name)": {
                "loss_name": {
                    "loss": torch.Tensor(...),
                    "weight": torch.Tensor(...)
                }
            }
        }
        """
        self.loss_values["total"] = 0
        # update loss value
        for loss_info in loss_dict.values():
            for loss_name, loss_data in loss_info.items():
                self.loss_values[loss_name] = loss_data["loss"]
                self.weight_values[loss_name] = loss_data.get("weight", 1.0)
        # compute total sum of losses
        for k in self.loss_values.keys():
            if k != "total":
                self.loss_values["total"] += self.loss_values[k] * self.weight_values[k]
        
        # average the values
        for k in self.loss_values.keys():
            self.avg_loss_list[k].append(float(self.loss_values[k]))

    def average_losses(self):
        for k in self.avg_loss_list.keys():
            sub_loss = torch.tensor(self.avg_loss_list[k])
            avg_sub_loss = torch.mean(sub_loss[~ torch.isnan(sub_loss)])
            self.avg_loss_values[k] = avg_sub_loss
        self.avg_loss_list = defaultdict(list)
        
    
    def get(self, key, default_value=None, is_avg_value=True):
        if is_avg_value:
            return self.avg_loss_values.get(key, default_value)
        else:
            return self.loss_values.get(key, default_value)

if __name__ == "__main__":
    test = LossHandler()
    test.zero_losses()


"""
This module is designed for recording:

1. support ``Recorder``. Class ``WithRecorder`` allows to use ``recorder`` to save statistics during training using csv and tensorboard.

"""

from .with_recorder import WithRecorder


import torch
import tarfile
from tqdm import tqdm
import numpy as np 
import logging
from pandas import read_csv
import os
import re
import shutil
from torchvision import datasets
from .base_dataset import BaseDataset

class ImageNet(BaseDataset):
    """
    `ImageNet <http://www.image-net.org/>`_
    """
    def init_dataset(self):
        self.root = os.path.join(self.root, "imagenet")
        if self.phase == "train":
            self.init_train_dataset()
        elif self.phase == "test":
            self.init_test_dataset()
        else:
            raise KeyError(
                "Invalid dataset phase: {}".format(
                    self.phase
                )
            )
    
    """
    train set
    """

    def init_train_dataset(self):
        self.img_src = os.path.join(self.root, "imagenet_train")
        self.img_folder = os.path.join(self.root, "train")
        if not os.path.exists(self.img_folder):
            logging.info("PID: {} - Start to extract train iamges: {}".format(
                os.getpid(),
                self.__class__.__name__
            ))
            self.extract_train_img(self.img_src, self.img_folder)
        self.dataset = datasets.ImageFolder(self.img_folder)
        self.labels = np.array([b for (a,b) in self.dataset.imgs])
        assert len(np.unique(self.labels)) == self.assert_num_classes
        assert self.__len__() == self.assert_num_samples
    
    def extract_train_img(self, img_src, img_dst):
        tar_list = os.listdir(img_src)
        for file_name in tqdm(tar_list):
            tar = tarfile.open(os.path.join(img_src, file_name), "r")
            tar.extractall(path=os.path.join(img_dst, file_name.split('.')[0]))
            tar.close()
        
    """
    val set (test)
    """

    def init_test_dataset(self):
        self.img_src = os.path.join(self.root, "imagenet_val")
        self.img_folder = os.path.join(self.root, "test")
        self.info_file_path = os.path.join(
            self.root, 
            "ILSVRC2012_devkit_t12", 
            "data", 
            "ILSVRC2012_validation_ground_truth.txt"
        )
        if not os.path.exists(self.img_folder):
            logging.info("PID: {} - Start to extract test images: {}".format(
                os.getpid(),
                self.__class__.__name__
            ))
            self.extract_test_img(self.img_src, self.img_folder)
        self.dataset = datasets.ImageFolder(self.img_folder)
        self.labels = np.array([b for (a,b) in self.dataset.imgs])
        assert len(np.unique(self.labels)) == self.assert_num_classes
        assert self.__len__() == self.assert_num_samples
    
    def extract_test_img(self, img_src, img_dst):
        img_name_list = os.listdir(img_src)
        img_name_dict = {
            int(re.findall("\d+", name)[1]): name
            for name in img_name_list
        }
        info_file = read_csv(
            self.info_file_path,
            delim_whitespace=True,
            header=None
        ).values
        img_label_dict = {
            int(idx+1): int(item)
            for idx, item in enumerate(info_file)
        }
        for k, curr_img_label in tqdm(img_label_dict.items()):
            curr_img_name = img_name_dict[k]
            img_label_folder = os.path.join(img_dst, str(curr_img_label))
            if not os.path.exists(img_label_folder):
                os.makedirs(img_label_folder)
            img_src_path = os.path.join(img_src, curr_img_name)
            img_dst_path = os.path.join(img_label_folder, curr_img_name)
            shutil.copyfile(img_src_path, img_dst_path)
            
    
    def get_labels(self):
        return self.labels 
    
    def __len__(self):
        return len(self.dataset)
    
    def get_image_label(self, idx):
        return self.dataset[idx]


"""
This module implements the evaluation indicators, such as F1-score, recall@k, precision@k and etc using faiss-gpu package.

- ``calculator.py``: A evaluation manager class.
- ``metrics.py``: All evaluation functions are implemented here.
"""

from .calculator import Calculator
from .calculator_mat import CalculatorFromMat


import torch
import torch.nn as nn
import math
from copy import deepcopy
import logging
import torch.distributed as dist 

from ..base_collector import BaseCollector
from ...misc import utils
from ...models import BatchNormMLP

class SimSiamCollector(BaseCollector):
    """
    Paper: `Exploring Simple Siamese Representation Learning <https://arxiv.org/abs/2011.10566>`_

    This method use none of the following to learn meaningful representations:

    1. negative sample pairs;
    2. large batches;
    3. momentum encoders.

    And a stop-gradient operation plays an essential role in preventing collapsing.
    """
    def __init__(
        self,
        *args,
        **kwargs
    ):
        super(SimSiamCollector, self).__init__(*args, **kwargs) 
        self.predictor = BatchNormMLP(
            layer_size_list=[2048, 512, 2048],
            relu_list=[True, False],
            bn_list=[True, False]
        )
    
    def forward(self, data, embeddings, labels) -> tuple:
        """
        For simplicity, two data streams will be combined together and be passed through ``embeddings`` parameter. In function ``collect``, two data streams will be split (first half for first stream; second half for second stream).

        Args:
            data (torch.Tensor):
                A batch of key images (**not used**). size: :math:`B \\times C \\times H \\times W`
            embeddings (torch.Tensor):
                A batch of query embeddings. size: :math:`2B \\times dim`
            labels (torch.Tensor):
                Labels of the input. size: :math:`2B \\times 1`
        """
        # split two streams
        N = embeddings.size(0)
        assert N % 2 == 0
        z1, z2 = embeddings[:N//2], embeddings[N//2:]
        labels = labels[:N//2]

        # compute p
        p1 = self.predictor(z1)
        p2 = self.predictor(z2)

        metric_mat = 0.5 * (
            self.metric(p1, z2.detach()) +
            self.metric(p2, z1.detach())
        )
        return (
            metric_mat,
            labels.unsqueeze(1),
            labels.unsqueeze(0),
            False
        )



import os

CLASS_KEY = "type"
PARAMS_KEY = "params"
INITIATE_KEY = "INITIATE"
WRAPPER_KEY = "WRAPPER"
OBJECTS_KEY = "objects"
DEFAULT_NAME = "default"
WILDCARD = "*"
SPLITER = "/"
ATTRSYM = "-"

LINK_GENERAL_KEY = "LINK_SETTING"
PIPELINE_KEY = "PIPELINE_SETTING"

def search_params_condition(params):
    if isinstance(params, list):
        if len(params) > 0:
            if isinstance(params[0], str):
                if SEARCH_PARAMS_FLAG in params[0]:
                    return params[0]
    return False

SEARCH_PARAMS_FLAG = "~~"
SEARCH_PARAMS_CONDITION = search_params_condition
SEARCH_TOP_CLASS = lambda x: x[1]
SEARCH_TAR_NAME = lambda x: x[2]

INITATE_ORDER_LIST = [
    "recorders",
    "metrics",
    "models",
    "collectors",
    "selectors",
    "losses",
    "evaluators",
    "optimizers",
    "schedulers",
    "gradclipper",
    "transforms",
    "datasets",
    "samplers",
    "collatefns",
    "trainers",
    "testers",
    "managers",
]

WRAPPER_LIST = [
    "models",
    "collectors",
    "selectors",
    "losses"
]


import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class MultiSimilarityLoss(BaseLoss):
    """
    paper: `Multi-Similarity Loss With General Pair Weighting for Deep Metric Learning <https://openaccess.thecvf.com/content_CVPR_2019/html/Wang_Multi-Similarity_Loss_With_General_Pair_Weighting_for_Deep_Metric_Learning_CVPR_2019_paper.html>`_
    """
    def __init__(
        self,
        alpha=2,
        beta=50,
        base=0.5,
        **kwargs
    ):
        super(MultiSimilarityLoss, self).__init__(**kwargs)
        self.alpha = alpha
        self.beta = beta
        self.base = base
        
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        # get masks
        pos_mask = row_labels == col_labels
        neg_mask = ~ pos_mask
        if is_same_source:
            pos_mask.fill_diagonal_(False)

        # compute loss
        pos_loss = (1.0 / self.alpha) * l_f.logsumexp(
            - self.alpha * (metric_mat - self.base), keep_mask=pos_mask, add_one=True, dim=1
        )
        neg_loss = (1.0 / self.beta) * l_f.logsumexp(
            self.beta * (metric_mat - self.base), keep_mask=neg_mask, add_one=True, dim=1
        )
        loss = torch.mean(pos_loss) + torch.mean(neg_loss)

        return loss


import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class SoftTripleLoss(BaseLoss):
    """
    paper: `SoftTriple Loss: Deep Metric Learning Without Triplet Sampling <https://openaccess.thecvf.com/content_ICCV_2019/html/Qian_SoftTriple_Loss_Deep_Metric_Learning_Without_Triplet_Sampling_ICCV_2019_paper.html>`_
    """
    def __init__(
        self,
        centers_per_class=10,
        num_classes=100,
        la=20,
        gamma=0.1,
        margin=0.01,
        **kwargs
    ):
        super(SoftTripleLoss, self).__init__(**kwargs)
        self.centers_per_class = centers_per_class
        self.num_classes = num_classes
        self.la = la 
        self.gamma = gamma
        self.margin = margin
    
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        metric_to_centers = metric_mat.view(
            -1, self.num_classes, self.centers_per_class
        )
        prob = torch.nn.functional.softmax(
            metric_to_centers * self.gamma, dim=2
        )
        metric_to_classes = torch.sum(
            prob * metric_to_centers, dim=2
        )
        margin = torch.zeros_like(metric_to_classes)
        margin[torch.arange(margin.shape[0]), row_labels.squeeze()] = self.margin
        loss = torch.nn.functional.cross_entropy(
            self.la * (metric_to_classes - margin), row_labels.squeeze()
        )
        return loss


import torch
import numpy as np 

from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

def get_unique_indices(indices_tuple, labels):
    a, p, _, _ = split_indices(indices_tuple)
    _, unique_idx = np.unique(labels[a].cpu().numpy(), return_index=True)
    return a[unique_idx], p[unique_idx]

def indices_to_pairs(metric_mat, indices_tuple, assert_shape=[2, 3], return_pn=False):
    tuples = indices_tuple[INDICES_TUPLE]
    assert tuples.shape[1] in assert_shape
    
    if tuples.shape[1] == 3:
        pos_pair = metric_mat[tuples[:,0], tuples[:,1]]
        neg_pair = metric_mat[tuples[:,0], tuples[:,2]]
        if return_pn:
            pn_pair = metric_mat[tuples[:,1], tuples[:,2]]
            return pos_pair, neg_pair, pn_pair
        else:
            return pos_pair, neg_pair

    elif tuples.shape[1] == 2:
        flags = indices_tuple[INDICES_FLAG]
        pos_indices = torch.where(flags)[0]
        neg_indices = torch.where(flags^1)[0]
        pos_pair = metric_mat[tuples[pos_indices,0], tuples[pos_indices,1]]
        neg_pair = metric_mat[tuples[neg_indices,0], tuples[neg_indices,1]]
        return pos_pair, neg_pair

    else:
        raise ValueError("Not supported tuple structure!")

def split_indices(indices_tuple):
    tuples = indices_tuple[INDICES_TUPLE]
    if tuples.shape[1] == 3:
        return tuples[:,0], tuples[:,1], tuples[:,0], tuples[:,2]
    elif tuples.shape[1] == 2:
        flags = indices_tuple[INDICES_FLAG]
        pos_indices = torch.where(flags)[0]
        neg_indices = torch.where(flags^1)[0]
        return tuples[pos_indices,0], tuples[pos_indices,1], tuples[neg_indices,0], tuples[neg_indices,1]
    else:
        raise ValueError("Not supported tuple structure!")

def sumexp(x, keep_mask=None, dim=1):
    x_exp = (
        torch.exp(x) * keep_mask
        if keep_mask is not None
        else torch.exp(x)
    )
    x_sumexp = torch.sum(x_exp, dim=dim, keepdim=True)
    return x_sumexp

def logsumexp(x, keep_mask=None, add_one=True, dim=1):
    x_exp = (
        torch.exp(x) * keep_mask
        if keep_mask is not None
        else torch.exp(x)
    )

    if add_one:
        ones = torch.ones(x_exp.size(dim - 1), dtype=x_exp.dtype, device=x_exp.device).unsqueeze(dim)
        x_exp = torch.cat([x_exp, ones], dim=dim)
    
    x_logsumexp = torch.log(
        torch.sum(x_exp, dim=dim, keepdim=True)
    )
    return x_logsumexp






import torch

from .base_distance_metric import BaseDistanceMetric

class SNRMetric(BaseDistanceMetric):
    def __call__(self, row_input, col_input=None):
        if col_input is None:
            col_input = row_input
        denominator = torch.var(row_input, dim=1)
        numerator = torch.var(row_input.unsqueeze(1) - col_input, dim=2)
        return numerator / (denominator.unsqueeze(1))


"""
There are two kinds of ``transforms``:

1. **augumentation method**. Define how an image is augumented.
2. **wrapping method**. Define how to combine multi streams.
"""

from .img_transforms import (
    ConvertToBGR,
    Multiplier
)
from .wrapper_transforms import (
    TwoCropsTransformWrapper,
    DefaultTransformWrapper
)
from torchvision.transforms import (
    Resize,
    RandomHorizontalFlip,
    RandomResizedCrop,
    RandomGrayscale,
    ColorJitter,
    CenterCrop,
    ToTensor,
    Normalize,
    Compose,
)


import torch
from torch.utils.data.sampler import Sampler
from collections import defaultdict
import numpy as np 
from copy import deepcopy

def safe_random_choice(input_list, size):
    replace = len(input_list) < size
    return np.random.choice(input_list, size=size, replace=replace)

class MPerClassFullSampler(Sampler):
    """
    Give m samples per class (Try to cover all the samples).

    Args:
        labels (np.ndarray):
            Ground truth of datasets
        m (int):
            M samples per class
        batch_size (int): 
            Batch size must be an interger multiple of m
    """
    def __init__(
        self,
        labels,
        m,
        batch_size,
    ):
        # raise NotImplementedError()
        self.labels = labels
        self.m = m 
        self.batch_size = batch_size

        self.init_indices()
    
    def init_indices(self):
        # construct indices
        self.labels_to_indices = defaultdict(list)
        for i, label in enumerate(self.labels):
            self.labels_to_indices[label].append(i)
        for k, v in self.labels_to_indices.items():
            self.labels_to_indices[k] = np.array(v).astype('int')
        self.labels_set = list(self.labels_to_indices.keys())
        self.min_label_num = min([len(v) for v in self.labels_to_indices.values()])
        self.list_size = self.min_label_num * len(self.labels_set)

        # assert
        assert self.list_size >= self.batch_size
        assert self.batch_size % self.m == 0
        self.list_size -= self.list_size % self.batch_size
        self.num_iters = self.list_size // self.batch_size
    
    def __len__(self):
        return self.list_size
        
    def __iter__(self):
        num_group = int(self.min_label_num / self.m)
        num_label = int(self.batch_size / self.m)

        idx_list = [0] * self.list_size
        store_dict = {
            label: safe_random_choice(self.labels_to_indices[label], self.min_label_num)
            for label in self.labels_set
        }
        store_dict = {
            k: [
                v[idx*self.m:(idx+1)*self.m]
                for idx in range(num_group)
            ]
            for k, v in store_dict.items()
        }
        self.labels_set = np.array(self.labels_set)
        total_label_list = []
        for _ in range(num_group):
            label_set = deepcopy(self.labels_set.tolist())
            np.random.shuffle(label_set)
            total_label_list += label_set

        list_idx = 0
        for i in range(self.num_iters):
            # avaliable curr-labels-set
            curr_labels_set = total_label_list[i*num_label : (i+1)*num_label]

            # get batch-size samples
            for label in curr_labels_set:
                curr_indices = store_dict[label].pop(0)
                idx_list[list_idx:list_idx+self.m] = curr_indices
                list_idx += self.m 
        return iter(idx_list)

if __name__ == "__main__":
    labels_num = np.random.choice([60, 61, 62, 63], size=(75,), replace=True)
    labels = []
    for i in range(75):
        labels += [i] * labels_num[i]
    sampler = MPerClassFullSampler(labels=labels, m=4, batch_size=32)
    iter_ = iter(sampler)
    pass


import torch
import scipy.special
import math

from ..base_loss import BaseLoss

class LargeMarginSoftmaxLoss(BaseLoss):
    """
    modified from: https://github.com/KevinMusgrave/pytorch-metric-learning

    paper: `Large-Margin Softmax Loss for Convolutional Neural Networks <https://www.jmlr.org/proceedings/papers/v48/liud16.pdf>`_
    """
    def __init__(
        self,
        margin=4,
        scale=1,
        **kwargs
    ):
        super(LargeMarginSoftmaxLoss, self).__init__(**kwargs)
        self.margin = margin
        self.scale = scale
        self.initiate_margin()

    def required_metric(self):
        return ["cosine"]
    
    def initiate_margin(self):
        self.margin = int(self.margin)
        self.max_n = self.margin // 2
        ## For the trigonometric multiple-angle formula ##
        self.n_range = torch.Tensor([
            n for n in range(0, self.max_n + 1)
        ])
        self.margin_choose_n = torch.Tensor([
            scipy.special.binom(self.margin, 2 * n) for n in self.n_range
        ])
        self.cos_powers = torch.Tensor([
            self.margin - (2 * n) for n in self.n_range
        ])
        self.alternating = torch.Tensor([
            (-1) ** n for n in self.n_range
        ])
    
    def get_cos_with_margin(self, cosine):
        cosine = cosine.unsqueeze(1)
        for attr in ["n_range", "margin_choose_n", "cos_powers", "alternating"]:
            setattr(self, attr, getattr(self, attr).to(cosine.device))
        cos_powered = cosine ** self.cos_powers
        sin_powered = (1 - cosine ** 2) ** self.n_range
        terms = (
            self.alternating * self.margin_choose_n * cos_powered * sin_powered
        )
        return torch.sum(terms, dim=1)
    
    def get_target_mask(self, metric_mat, labels):
        batch_size = labels.size(0)
        mask = torch.zeros_like(metric_mat, device=metric_mat.device)
        mask[torch.arange(batch_size), labels] = 1
        return mask
    
    def get_angles(self, cosine_of_target_classes):
        angles = torch.acos(torch.clamp(cosine_of_target_classes, -1, 1))
        return angles
    
    def modify_cosine_of_target_classes(self, cosine_of_target_classes):
        cos_with_margin = self.get_cos_with_margin(cosine_of_target_classes)
        angles = self.get_angles(cosine_of_target_classes)
        with torch.no_grad():
            k = (
                angles / (math.pi / self.margin)
            ).floor()
        return ((-1) ** k) * cos_with_margin - (2 * k)
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        is_same_source=False,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        dtype, device = metric_mat.dtype, metric_mat.device
        mask = self.get_target_mask(metric_mat, row_labels.squeeze())
        cosine_of_target_classes = metric_mat[mask == 1]
        modified_cosine_of_target_classes = self.modify_cosine_of_target_classes(
            cosine_of_target_classes 
        )
        diff = (modified_cosine_of_target_classes - cosine_of_target_classes).unsqueeze(1)
        logits = metric_mat + (mask * diff)
        # TODO: lack scaling function
        loss = torch.nn.functional.cross_entropy(logits * self.scale, row_labels.squeeze())

        return loss
    

    



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gedml",
    version="0.1.1",
    author="Borui Zhang",
    author_email="zhang-br21@mails.tsinghua.edu.cn",
    description="GeDML is an easy-to-use generalized deep metric learning library, \
        which contains state-of-the-art deep metric learning algorithms and \
        auxiliary modules to build end-to-end compute vision systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zbr17/GeDML",
    project_urls={
        "Docs": "https://zbr17.github.io/GeDML/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "torch == 1.7.0",
        "faiss-gpu == 1.5.3",
        "torchvision == 0.8.0",
        "pretrainedmodels == 0.7.4",
        "libtmux >= 0.8.5",
        "numpy >= 1.19.2",
        "tqdm >= 4.59.0",
        "pandas >= 1.1.3",
        "scipy >= 1.5.4",
        "scikit-learn >= 0.24.1",
        "tensorboard >= 2.4.0",
        "wandb >= 0.10.22",
    ]
)


import logging
import os

os.environ["MASTER_ADDR"] = "localhost"
os.environ["MASTER_PORT"] = "10002"

import torch
import torch.distributed as dist 
import torch.backends.cudnn as cudnn
torch.multiprocessing.set_sharing_strategy('file_system')

from ..creators import ConfigHandler
from ..misc import utils

# TODO: to be perfected

class DistributedRunner:
    def __init__(
        self,
        opt
    ):
        self.opt = opt
        self.assert_opt()
    
    @property
    def required_members_list(self):
        return ["gpu", "world_size", "rank", "batch_size", "num_workers"]
    
    def assert_opt(self):
        # assert
        assert all(
            [
                getattr(self.opt, item, None) is not None
                for item in self.required_members_list
            ]
        )

    def run(self):
        cudnn.deterministic = True
        cudnn.benchmark = True
        # initiate torch.distributed
        dist.init_process_group(
            backend='nccl',
            init_method='env://',
            world_size=self.opt.world_size,
            rank=self.opt.rank
        )

        # get config_handler
        config_handler = ConfigHandler(
            link_path=self.opt.link_path
        )
        del self.opt.link_path

        # change link config
        if self.opt.dataset == "CUB200":
            config_handler.link_config["datasets"] = [
                {"train": "cub200_train.yaml"},
                {"test": "cub200_test.yaml"}
            ]
        elif self.opt.dataset == "ImageNet":
            config_handler.link_config["datasets"] = [
                {"train": "imagenet_train.yaml"},
                {"test": "imagenet_test.yaml"}
            ]
        del self.opt.dataset
        
        # initiate params_dict
        config_handler.get_params_dict()

        # modify parameters: save_path, gpu_device
        save_path = config_handler.get_certain_params_dict(["save_path"])
        save_path = utils.get_first_dict_value(save_path["save_path"])
        save_path = os.path.join(
            save_path,
            "subprocess_rank_{}".format(self.opt.rank)
        )
        del self.opt.rank

        modify_dict = {
            "device": self.opt.gpu,
            "is_distributed": True
        }
        del self.opt.gpu
        self.opt = self.opt.__dict__
        modify_dict.update(self.opt)
        objects_dict = config_handler.create_all(modify_dict)
        manager = utils.get_default(objects_dict, "managers")
        manager.run()



import torch
from torch.utils.data.sampler import Sampler
from collections import defaultdict
import numpy as np 

def safe_random_choice(input_list, size):
    replace = len(input_list) < size
    return np.random.choice(input_list, size=size, replace=replace)

class MPerClassSampler(Sampler):
    """
    Give m samples per class.

    Args:
        labels (np.ndarray):
            Ground truth of datasets
        m (int):
            M samples per class
        batch_size (int): 
            Batch size must be an interger multiple of m
    """
    def __init__(
        self,
        labels,
        m,
        batch_size,
    ):
        # raise NotImplementedError()
        self.labels = labels
        self.m = m 
        self.batch_size = batch_size

        self.init_indices()
    
    def init_indices(self):
        # construct indices
        self.labels_to_indices = defaultdict(list)
        for i, label in enumerate(self.labels):
            self.labels_to_indices[label].append(i)
        for k, v in self.labels_to_indices.items():
            self.labels_to_indices[k] = np.array(v).astype('int')
        self.labels_set = list(self.labels_to_indices.keys())
        self.list_size = self.m * len(self.labels_set)

        # assert
        assert self.list_size >= self.batch_size
        assert self.batch_size % self.m == 0
        self.list_size -= self.list_size % self.batch_size
        self.num_iters = self.list_size // self.batch_size
    
    def __len__(self):
        return self.list_size
        
    def __iter__(self):
        idx_list = [0] * self.list_size
        i = 0
        for _ in range(self.num_iters):
            np.random.shuffle(self.labels_set)
            curr_labels_set = self.labels_set[:self.batch_size // self.m]
            for label in curr_labels_set:
                curr_indices = self.labels_to_indices[label]
                idx_list[i:i+self.m] = safe_random_choice(curr_indices, size=self.m)
                i += self.m 
        return iter(idx_list)



import torch

from .base_similarity_metric import BaseSimilarityMetric

class CosineMetric(BaseSimilarityMetric):
    def __call__(self, row_input, col_input=None):
        if col_input is None:
            col_input = row_input
        return torch.matmul(row_input, col_input.t())
         


# Modified from github: [facebookresearch / deit](https://github.com/facebookresearch/deit)

import torch
import torch.nn as nn
from functools import partial

from timm.models.vision_transformer import VisionTransformer, _cfg
from timm.models.registry import register_model
from timm.models.layers import trunc_normal_


__all__ = [
    'deit_tiny_patch16_224', 'deit_small_patch16_224', 'deit_base_patch16_224',
    'deit_tiny_distilled_patch16_224', 'deit_small_distilled_patch16_224',
    'deit_base_distilled_patch16_224', 'deit_base_patch16_384',
    'deit_base_distilled_patch16_384',
]

class Identity(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, x):
        return x

class DistilledVisionTransformer(VisionTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dist_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
        num_patches = self.patch_embed.num_patches
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 2, self.embed_dim))
        self.head_dist = nn.Linear(self.embed_dim, self.num_classes) if self.num_classes > 0 else nn.Identity()

        trunc_normal_(self.dist_token, std=.02)
        trunc_normal_(self.pos_embed, std=.02)
        self.head_dist.apply(self._init_weights)

    def forward_features(self, x):
        # taken from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
        # with slight modifications to add the dist_token
        B = x.shape[0]
        x = self.patch_embed(x)

        cls_tokens = self.cls_token.expand(B, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
        dist_token = self.dist_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, dist_token, x), dim=1)

        x = x + self.pos_embed
        x = self.pos_drop(x)

        for blk in self.blocks:
            x = blk(x)

        x = self.norm(x)
        return x[:, 0], x[:, 1]

    def forward(self, x):
        x, x_dist = self.forward_features(x)
        x = self.head(x)
        x_dist = self.head_dist(x_dist)
        if self.training:
            return x, x_dist
        else:
            # during inference, return the average of both classifier predictions
            return (x + x_dist) / 2


@register_model
def deit_tiny_patch16_224(pretrained=False, **kwargs):
    model = VisionTransformer(
        patch_size=16, embed_dim=192, depth=12, num_heads=3, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_tiny_patch16_224-a1311bcf.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model

@register_model
def deit_small_patch16_224(pretrained=False, **kwargs):
    model = VisionTransformer(
        patch_size=16, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_small_patch16_224-cd65a155.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    # change the name and delete the last linear
    model.last_linear = model.head
    model.head = Identity()
    return model


@register_model
def deit_base_patch16_224(pretrained=False, **kwargs):
    model = VisionTransformer(
        patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_base_patch16_224-b5f2ef4d.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


@register_model
def deit_tiny_distilled_patch16_224(pretrained=False, **kwargs):
    model = DistilledVisionTransformer(
        patch_size=16, embed_dim=192, depth=12, num_heads=3, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_tiny_distilled_patch16_224-b40b3cf7.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


@register_model
def deit_small_distilled_patch16_224(pretrained=False, **kwargs):
    model = DistilledVisionTransformer(
        patch_size=16, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_small_distilled_patch16_224-649709d9.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


@register_model
def deit_base_distilled_patch16_224(pretrained=False, **kwargs):
    model = DistilledVisionTransformer(
        patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_base_distilled_patch16_224-df68dfff.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


@register_model
def deit_base_patch16_384(pretrained=False, **kwargs):
    model = VisionTransformer(
        img_size=384, patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_base_patch16_384-8de9b5d1.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


@register_model
def deit_base_distilled_patch16_384(pretrained=False, **kwargs):
    model = DistilledVisionTransformer(
        img_size=384, patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    model.default_cfg = _cfg()
    if pretrained:
        checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_base_distilled_patch16_384-d0272ac0.pth",
            map_location="cpu", check_hash=True
        )
        model.load_state_dict(checkpoint["model"])
    return model


from .base_creator import BaseCreator

class lossesCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        from ....core import losses
        self.package = [losses]


import numpy as np 
import torch

from .base_selector import BaseSelector
from ...config.setting.core_setting import (
    INDICES_TUPLE,
    INDICES_FLAG
)

class RandomTripletSelector(BaseSelector):
    """
    Semi-hard sampling method, euclidean distance metric is required.
    """
    def __init__(self, **kwargs):
        super(BaseSelector, self).__init__(**kwargs)
    
    def forward(
        self,
        metric_mat,
        row_labels,
        col_labels,
        is_same_source=False
    ):
        """
        Randomly select a positive sample and a negative sample
        """
        device = metric_mat.device
        bs = metric_mat.size(0)
        # pos and neg mask
        matches = (row_labels == col_labels).byte()
        diffs = matches ^ 1
        if is_same_source:
            matches.fill_diagonal_(0)

        has_pn_mask = torch.where(
            (torch.sum(matches, dim=-1) > 0) & 
            (torch.sum(diffs, dim=-1) > 0)
        )[0]

        a_ids = torch.arange(bs)[has_pn_mask].to(device)
        # select positive samples
        p_ids = torch.multinomial(
            input=matches.float()[has_pn_mask, :],
            num_samples=1,
            replacement=True
        ).flatten()

        # select negative samples
        n_ids = torch.multinomial(
            input=diffs.float()[has_pn_mask, :],
            num_samples=1,
            replacement=True
        ).flatten()

        indices_tuple = {
            INDICES_TUPLE: torch.stack([a_ids, p_ids, n_ids], dim=1),
            INDICES_FLAG: None
        }
        weight = None
        return (
            metric_mat,
            row_labels,
            col_labels,
            is_same_source,
            indices_tuple,
            weight,
        )


import torch
import numpy as np 
import math

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class AngularLoss(BaseLoss):
    """
    paper: `Deep Metric Learning with Angular Loss <https://openaccess.thecvf.com/content_iccv_2017/html/Wang_Deep_Metric_Learning_ICCV_2017_paper.html>`_
    """
    def __init__(
        self,
        alpha=40,
        **kwargs
    ):
        super(AngularLoss, self).__init__(**kwargs)
        self.alpha = alpha
        self.sq_tan_alpha = math.tan(math.radians(self.alpha)) ** 2
    
    def required_metric(self):
        return ["cosine"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple,
        is_same_source=False,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        # get indices
        anchor_index, positive_index, keep_mask = self.get_pairs(row_labels, col_labels, indices_tuple)

        # compute loss
        ap_pair = metric_mat[anchor_index, positive_index].unsqueeze(1)
        ap_n_pair = (
            metric_mat[anchor_index,:] +
            metric_mat[positive_index,:]
        )
        f_apn = (
            (4 * self.sq_tan_alpha * ap_n_pair) -
            (2 * (1 + self.sq_tan_alpha) * ap_pair)
        )
        loss = l_f.logsumexp(f_apn, keep_mask=keep_mask, add_one=True, dim=1)
        loss = torch.mean(loss)

        return loss
    
    def get_pairs(self, row_labels, col_labels, indices_tuple):
        a1, p, a2, _ = l_f.split_indices(indices_tuple)
        if len(a1) == 0 or len(a2) == 0:
            return [None] * 4
        keep_mask = row_labels[a1, :] != col_labels
        return a1, p, keep_mask.byte()



from .mlp import MLP
from .mlp import BatchNormMLP
from .identity import Identity
from .resnet50 import resnet50
from .deit import deit_small_patch16_224 as DeiT_S
# from pretrainedmodels import bninception
from .bninception import bninception
from torchvision.models import googlenet


from .base_creator import BaseCreator
from torch import optim
import torch
import logging

class optimizersCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [optim]
    
    def maybe_modify_args(self, module_args):
        if isinstance(module_args["params"], torch.nn.Module):
            parameters = list(filter(lambda p: p.requires_grad, module_args['params'].parameters()))
            logging.info(
                "Optimizer params from: {} id: {}".format(
                    module_args["params"].__class__.__name__,
                    id(module_args["params"])
                )
            )
        else:
            parameters = module_args["params"]
            logging.info(
                "Optimizer params from: {} id: {}".format(
                    str(parameters), id(parameters)
                )
            )
        module_args['params'] = parameters
        return module_args


from .base_creator import BaseCreator
from ....core import samplers

class samplersCreator(BaseCreator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def prepare_packages(self):
        self.package = [samplers]
    
    def maybe_modify_args(self, module_args):
        module_args["labels"] = module_args["labels"].get_labels()
        return module_args


import numpy as np 
import torch
from ..config.setting.recorder_setting import TO_RECORD_LIST
from . import utils

class BaseInfoGetter:
    def __init__(self):
        self.dict_counts = {}
    
    def assert_from_obj(self, obj):
        to_record_list = getattr(obj, TO_RECORD_LIST, None)
        assert to_record_list is not None, 'obj must have attribute: {}!'.format(TO_RECORD_LIST)
        return to_record_list
    
    def convert_from_obj(self, obj, name=None):
        name = obj.__class__.__name__ if name is None else name
        output = {}
        to_record_list = self.assert_from_obj(obj)
        for record_item in to_record_list:
            output[record_item] = utils.get_value(getattr(obj, record_item, 0))
        return {
            self.get_description(name): output
        }

    def get_data(self, data: (dict or ...), name=None) -> (dict, int):
        """
        This is the main API for BaseInfoGetter class.

        Args:
            data        (dict or obj):      Dict or obj (has TO_RECORD_LIST attribute)
            name        (str)
        Returns:
            [0]         (dict):             Data to be recorded
            [1]         (int):              Step
        """
        # extract data to dict
        output = self.convert_from_obj(data, name) if not isinstance(data, dict) else data
        assert len(list(output.keys())) < 2, 'first-level keys must be less than 2 in output dict'
        # increase counts
        primary_key = utils.get_first_key_of_dict(output)
        self.increase_counts(primary_key)
        return output, self.dict_counts[primary_key]
    
    def increase_counts(self, key):
        if key not in self.dict_counts:
            self.dict_counts[key] = 0 
        else:
            self.dict_counts[key] += 1
        
    def get_description(self, name, spliter='-'):
        description = spliter.join([name, 'attribute'])
        return description

    def get_value(self, data):
        if isinstance(data, torch.Tensor):
            return data.item()
        elif isinstance(data, (int, float)):
            return data
        else:
            raise TypeError("Data must be torch.Tensor, int, or float!")



import torch

from ...misc import loss_function as l_f 
from ..base_loss import BaseLoss

class LiftedStructureLoss(BaseLoss):
    """
    paper: `Deep Metric Learning via Lifted Structured Feature Embedding <https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/Song_Deep_Metric_Learning_CVPR_2016_paper.html>`_
    """
    def __init__(
        self,
        neg_margin=1,
        pos_margin=0,
        **kwargs
    ):
        super(LiftedStructureLoss, self).__init__(**kwargs)
        self.neg_margin = neg_margin
        self.pos_margin = pos_margin
    
    def required_metric(self):
        return ["euclid"]
    
    def compute_loss(
        self,
        metric_mat,
        row_labels,
        col_labels,
        indices_tuple=None,
        weights=None,
        is_same_source=False,
    ) -> torch.Tensor:
        a1, p, a2, _ = l_f.split_indices(indices_tuple)
        pos_pair, neg_pair = l_f.indices_to_pairs(metric_mat, indices_tuple)
        dtype = metric_mat.dtype

        if len(a1) > 0 and len(a2) > 0:
            pos_pair = pos_pair.unsqueeze(1)
            n_per_p = (
                (a2.unsqueeze(0) == a1.unsqueeze(1))
                | (a2.unsqueeze(0) == p.unsqueeze(1))
            ).type(dtype)
            neg_pair = neg_pair * n_per_p
            keep_mask = ~ (n_per_p == 0)

            remaining_neg_margin = self.neg_margin - neg_pair
            remaining_pos_margin = pos_pair - self.pos_margin
            neg_pair_loss = l_f.logsumexp(
                remaining_neg_margin, keep_mask=keep_mask, add_one=False, dim=1
            )
            loss_per_pos_pair = neg_pair_loss + remaining_pos_margin
            loss_per_pos_pair = torch.relu(loss_per_pos_pair) ** 2
            loss_per_pos_pair /= 2
            loss = torch.mean(loss_per_pos_pair)
            return loss
        else:
            return torch.sum(metric_mat * 0)



import torch
from abc import ABCMeta, abstractmethod

from ..modules import WithRecorder

class BaseCollector(WithRecorder, metaclass=ABCMeta):
    """
    Base class of collector module, which defines main collector method in function ``collect`` and ``update``, and defines default parameters in function ``output_list``, ``input_list`` and ``_default_next_module``.

    Args:
        metric (metric instance):
            metric to compute matrix (e.g. euclidean or cosine)
    
    Example:
        >>> metric = MetricFactory(is_normalize=True, metric_name="cosine")
        >>> data = torch.randn(10, 3, 227, 227)
        >>> embeddings = torch.randn(10, 128)
        >>> labels = torch.randint(0, 3, size=(10,))
        >>> collector = DefaultCollector(metric=metric)
        >>> # collector forward
        >>> output_dict = collector(data, embeddings, labels)
    """
    def __init__(self, metric, **kwargs):
        super().__init__(**kwargs)
        self.metric = metric
    
    @property
    def is_global_collector(self):
        return False
    
    def update(self, *args, **kwargs):
        """
        Define the interface that collector can update itself by giving specific information (default do nothing)
        """
        pass

    def forward(
        self, 
        data, 
        embeddings, 
        labels
    ) -> tuple:
        """
        In ``collect`` function, three kinds of operation may be done:

        1. maintain sets of parameters about collecting (or synthesizing) samples
        2. compute metric matrix and pass to next module
        3. compute some regularization term using embeddings

        Args:
            data (torch.Tensor):
                Images with RGB channels. size: :math:`B \\times C \\times H \\times W`
            embeddings (torch.Tensor):
                Embedding. size: :math:`B \\times dim`
            lables (torch.Tensor): 
                Ground truth of dataset. size: :math:`B \\times 1`

        Returns:
            tuple: include metric matrix, labels etc according to function ``output_list``.

            Let :math:`B_{row}` be the length of rows and :math:`B_{col}` be the length of columns, typical output type is listed below:

            1. metric matrix (torch.Tensor): size: :math:`B_{row} \\times B_{col}`
            2. labels of rows (torch.Tensor): size: :math:`B_{row} \\times 1` or :math:`B_{row} \\times B_{col}`
            3. labels of columns (torch.Tensor): size: :math:`1 \\times B_{col}` or :math:`B_{row} \\times B_{col}`
            4. is_from_same_source (bool): indicate whether row vectors and column vectors are from the same data
        
        """
        metric_mat, labels, proxies_labels, is_same_source = None, None, None, False
        return (
            metric_mat, 
            labels, 
            proxies_labels, 
            is_same_source
        )



import torch
import logging
from tqdm import tqdm
import torch.nn.functional as F 

from .. import ProxyCollector
from ._default_global_collector import _DefaultGlobalCollector

class GlobalProxyCollector(ProxyCollector, _DefaultGlobalCollector):
    """
    Compute the global proxies before updating other parameters.
    """

    def __init__(
        self,
        optimizer_name="Adam",
        optimizer_param={"lr": 0.001},
        dataloader_param={"batch_size": 120, "drop_last": False, "shuffle": True, "num_workers": 8},
        max_iter=50000,
        error_bound=1e-3,
        total_patience=10,
        auth_weight=1.0,
        repre_weight=1.0,
        disc_weight=1.0,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.optimizer_name = optimizer_name
        self.optimizer_param = optimizer_param
        self.dataloader_param = dataloader_param
        self.max_iter = max_iter
        self.error_bound = error_bound
        self.total_patience = total_patience
        self.auth_weight = auth_weight
        self.repre_weight = repre_weight
        self.disc_weight = disc_weight

        self.datasets, self.models = None, None

    def global_update(self, trainer):
        logging.info("Start to compute global embeddings")
        self.prepare(trainer)
        self.compute_global_embedding()
        self.compute_global_proxy()
    
    def prepare(self, trainer):
        _DefaultGlobalCollector.prepare(self, trainer)

        # get optimizer
        self.optimizer_param["params"] = [self.proxies]
        self.optimizer = getattr(
            torch.optim, 
            self.optimizer_name,
        )(**self.optimizer_param)
    
    def compute_global_proxy(self):
        pre_loss_value = -100
        patience_count = 0
        pbar = tqdm()
        index = 0
        while(True):
            # update proxies
            index += 1
            self.optimizer.zero_grad()
            cur_loss = self.loss_func(self.embeddings, self.labels)
            cur_loss.backward()
            self.optimizer.step()

            # raise error when get max_iter
            if index > self.max_iter:
                logging.warn("Reach MAX_ITER!")
                raise RuntimeError()

            # termination condition
            cur_error = abs(cur_loss.item() - pre_loss_value)
            pbar.set_description("CurrError={}, index={}".format(
                cur_error, index
            ))
            pre_loss_value = cur_loss.item()
            if cur_error < self.error_bound:
                patience_count += 1
                if patience_count > self.total_patience:
                    pbar.close()
                    logging.info("Have reached the solution of proxies!")
                    break
            else:
                patience_count = 0

    def loss_func(self, embeddings, labels):
        dtype = embeddings.dtype
        # normalize
        embeddings = F.normalize(embeddings, dim=-1, p=2)
        proxies = F.normalize(self.proxies, dim=-1, p=2)

        # compute cosine matrix
        metric_mat = torch.matmul(embeddings, proxies.t())
        pos_mask = (labels.unsqueeze(1) == self.proxy_labels.unsqueeze(0)).byte()

        # compute loss-auth
        masked_metric_mat = metric_mat * pos_mask
        masked_metric_mat[pos_mask == 0] = torch.finfo(dtype).min
        loss_auth = torch.mean(torch.max(masked_metric_mat, dim=0)[0])

        # compute loss-repre
        loss_repre = torch.mean(torch.max(masked_metric_mat, dim=1)[0])

        if self.centers_per_class > 1 and self.disc_weight > 0:
            # compute loss-disc
            proxy_metric_mat = torch.matmul(proxies, proxies.t())
            proxy_pos_mask = (self.proxy_labels.unsqueeze(1) == self.proxy_labels.unsqueeze(0)).byte()
            proxy_neg_mask = proxy_pos_mask ^ 1
            proxy_pos_mask.fill_diagonal_(0)
            
            ## compute pos metric
            pos_sum_exp = torch.sum(
                torch.exp(
                    - proxy_metric_mat
                ) * proxy_pos_mask, dim=-1
            )
            neg_sum_exp = torch.sum(
                torch.exp(
                    proxy_metric_mat
                ) * proxy_neg_mask, dim=-1
            )
            loss_disc = torch.mean(
                torch.log(
                    1 + pos_sum_exp * neg_sum_exp
                )
            )
        else:
            loss_disc = 0
        
        return (
            - self.auth_weight * loss_auth -
            self.repre_weight * loss_repre + 
            self.disc_weight * loss_disc
        )

    